# -*- coding: utf-8 -*-
"""
Paasify Application library

This library provides a convenient paasify user friendly API.

Example:
``` py title="test.py"
from paasify.app import PaasifyApp

app = PaasifyApp()

print (app.info())
prj = app.load_project()
prj.dump()
```

"""

# Imports
# =====================================================================

import os
from pprint import pprint  # noqa: F401

from cafram.utils import (
    to_json,
    _exec,
    write_file,
)
from cafram.nodes import NodeMap

import paasify.errors as error
from paasify.common import ensure_dir_exists, get_paasify_pkg_dir
from paasify.framework import PaasifyObj, FileLookup
from paasify.projects import PaasifyProject, PaasifyProjectConfig
from paasify.stacks import StackManager
from paasify.sources import SourcesManager
from paasify.collections import CollectionDocumator, gen_schema_doc


# Main Application class
# =====================================================================


class PaasifyApp(NodeMap, PaasifyObj):
    "Paasify Main application Instance"

    ident = "Paasify App"

    conf_default = {
        "config": {},
        "project": {},
    }

    conf_children = [
        {
            "key": "project",
            "cls": PaasifyProject,
            "action": "unset",
        },
    ]

    conf_schema = {
        # "$defs": {
        #     "AppProject": PaasifyProject.conf_schema,
        # },
        # "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "title": "Paasify App",
        "description": "Paasify app implementation",
        "additionalProperties": False,
        # "required": [
        #     "stacks"
        # ],
        "default": {},
        "properties": {
            "project": {
                "title": "Project configuration",
                "description": "See: schema prj",
                # "oneOf": [
                #     {
                #         "$ref": "#/$defs/AppProject",
                #         "description": "Instanciate project",
                #         "type": "object",
                #     },
                #     {
                #         "description": "Config file or path",
                #         "type": "string",
                #     },
                #     {
                #         "description": "Do not instanciate project",
                #         "type": "null",
                #     },
                # ],
            },
            "config": {
                "title": "Application configuration",
                "type": "object",
                "additionalProperties": True,
                # TODO: Add schema for appp
            },
        },
    }

    # Project management
    # ======================

    def load_project(self, path=None):
        "Return closest project"

        if self.project is not None:
            return self.project

        payload = path or {
            "_runtime": self.config,
        }

        prj = PaasifyProject(
            parent=self,
            payload=payload,  # Only string or nested runtime dict
        )

        self.add_child("project", prj)
        return prj

    def new_project(self, path, source=None):
        """Create a new project"""

        assets_dir = get_paasify_pkg_dir()
        changed = ensure_dir_exists(path)

        # Prepare configs
        build_conf = {
            "paasify": {
                "dest": os.path.join(path, "paasify.yml"),
            },
            "gitignore": {
                "dest": os.path.join(path, ".gitignore"),
            },
        }
        lookups = {
            "paasify": {
                "path": os.path.join(assets_dir, "assets"),
                "pattern": ["paasify.yml", "paasify.yaml"],
            },
            "gitignore": {
                "path": os.path.join(assets_dir, "assets"),
                "pattern": "gitignore",
            },
        }

        # Add source if requested
        if source:
            sources = {
                "paasify": {"path": source, "pattern": ["paasify.yml", "paasify.yaml"]},
                "gitignore": {
                    "path": source,
                    "pattern": ".gitignore",
                },
            }

        # Build each assets
        for name, conf in lookups.items():
            lookup = FileLookup()
            if source:
                lookup.append(sources[name]["path"], sources[name]["pattern"])
            lookup.append(conf["path"], conf["pattern"])
            match = lookup.match(first=True)
            if not match:
                continue

            # Copy files
            src = match["match"]
            dest = build_conf[name]["dest"]
            if not os.path.exists(dest):
                self.log.notice(f"Creating: '{dest}' from '{src}'")
                _exec("cp", cli_args=[src, dest])
                changed = True
            else:
                self.log.info(f"Skip: '{dest}' as it already exists")

        msg = "No changes in existing project"
        if changed:
            if source:
                msg = f"New project updated in: {path} from {source}"
            else:
                msg = f"New project updated in: {path}"
        self.log.notice(msg)

    # App commands
    # ======================

    def info(self, autoload=None):
        """Report app config"""

        print("Paasify App Info:")
        print("==================")
        for key, val in self.config.items():
            print(f"  {key}: {val}")

        print("\nPaasify Project Info:")
        print("==================")

        # Autoload default project
        msg = ""
        if autoload is None or bool(autoload):
            try:
                if not self.project:
                    self.log.notice("Info is autoloading project")
                    self.load_project()
            except error.ProjectNotFound as err:
                msg = err
                if autoload is True:
                    raise error.ProjectNotFound(err) from err

        if self.project:
            # Report with active project if available
            for key, val in self.project.runtime.get_value(lvl=-1).items():
                print(f"  {key}: {val}")
        else:
            print(f"  {msg}")

    def cmd_document(self, path=None, dest_dir=None, mkdocs_config=None):
        "Generate documentation"

        documator = CollectionDocumator(
            parent=self, path=path, dest_dir=dest_dir, mkdocs_config=mkdocs_config
        )
        documator.generate(
            mkdocs_config=None,
        )

    # pylint: disable=redefined-builtin
    def cmd_config_schema(self, format=None, dest_dir=None):
        """Returns the configuration json schema

        Args:
            dest_dir (str, optional): Destination directory

        """

        config = dict(
            minify=False,
            deprecated_from_description=True,
            default_from_description=False,
            expand_buttons=False,
            link_to_reused_ref=False,
            copy_css=True,
            copy_js=True,
            # config=config_file,
            # config_parameters=config,
        )
        # config = {}
        self.log.info("TEST 1")

        # Ensure that dest dir exists
        if dest_dir:
            self.log.notice(f"Documentation directory: {dest_dir}")
            if not os.path.isdir(dest_dir):
                self.log.notice(f"Creating parent directories: {dest_dir}")
                os.makedirs(dest_dir)

        targets = ["app", "prj", "prj_config", "prj_sources", "prj_stacks"]
        ret = {}
        for target in targets:

            # Select target to document
            if target == "app":
                schema = self.conf_schema
            elif target == "prj":
                schema = PaasifyProject.conf_schema
            elif target == "prj_config":
                schema = PaasifyProjectConfig.conf_schema
            elif target == "prj_sources":
                schema = SourcesManager.conf_schema
            elif target == "prj_stacks":
                schema = StackManager.conf_schema
            else:
                raise NotImplementedError(f"Target '{target}' is not supported")

            ret[target] = schema
            if not schema:
                assert False, "You found a bug!"

            if not dest_dir:
                continue

            slug = f"conf_{target}"
            json_file = os.path.join(dest_dir, f"{slug}.json")

            write_file(json_file, to_json(schema))
            self.log.info(f"Generate '{slug}.html' from: {json_file}")
            gen_schema_doc(json_file, dest_dir, f"{slug}.html", config=config)

        return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Paasify CLI interface

This API provides a similar experience as the CLI, but in Python.

Example:
``` py title="test.py"
from paasify.cli import app

paasify = app()
paasify.info()
paasify.apply()
```

"""

# pylint: disable=logging-fstring-interpolation

import os
import sys
import traceback
import logging
from enum import Enum
from typing import Optional
from pathlib import Path
from pprint import pprint  # noqa: F401

import yaml
import sh
import typer

from cafram.utils import get_logger, to_yaml, to_json
from cafram.base import CaframException

from paasify import __version__
import paasify.errors as error
from paasify.common import to_bool, OutputFormat
from paasify.app import PaasifyApp

# from rich.console import Console
# from rich.syntax import Syntax


# log = logging.getLogger()
log = logging.getLogger("paasify")

NOTTY = to_bool(os.environ.get("PAASIFY_NOTTY", os.environ.get("NOTTY", "False")))
PAASIFY_TRACE = to_bool(os.environ.get("PAASIFY_TRACE", False))
EDITOR = os.environ.get("EDITOR", "nano")


# CLI documenation
# ==============================

HELP_PROJECT_CMD = ""
HELP_STACKS_CMD = "Stacks Commands"
HELP_DOC_CMD = "Document Commands"
HELP_STACKS_HELPERS = "Stacks Commands"
HELP_SRC_CMD = "Sources Commands"

HELP_HEADER = f"""
Paasify - build your compose-files with ease

version: {__version__}

Paasify is a tool that build and deploy [bold]docker-compose.yml[/bold] files from a central
configuration file called [bold]paasify.yml[/bold]. In this file, define collections sources
and select your app. Then deploy your stacks, test, review and commit your
changes (and reiterate).

Full documentation: [italic green][bold]https://barbu-it.github.io/paasify/[/bold][/italic green]

🧭 Quickstart:

    To start a new project call [bold]my_prj[/bold]:
        paasify init my_prj

    Then add sources and stacks:
        {EDITOR} my_prj/paasify.yml

    Once you're ready, deploy your stacks:
        paasify -c my_prj apply


💡 Getting help:

    For more detailed information on each commands, please use the [bold]--help[/bold] flag
    Some commands provide an [bold]--explain[/bold] flag
    Extra logging is available, use [bold]-v[/bold] flags to increase verbosity


📢 Community:

    Project website:                [italic]https://github.com/barbu-it/paasify[/italic]
    Ask questions or report a bug:  [italic]https://github.com/barbu-it/paasify/issues[/italic]
    Original author: mrjk           [italic]https://github.com/mrjk[/italic]
    License: GPLv3

"""

# Exception handling
# ==============================


def test_logging():
    "Function to test logging"

    log.critical("SHOW CRITICAL")
    log.error("SHOW ERROR")
    log.warning("SHOW WARNING")
    log.notice("SHOW NOTICE")
    log.info("SHOW INFO")
    log.exec("SHOW EXEC")
    log.debug("SHOW DEBUG")
    log.trace("SHOW TRACE")


def clean_terminate(err):
    "Terminate nicely the program depending the exception"

    if PAASIFY_TRACE:
        log.error(traceback.format_exc())

    oserrors = [
        PermissionError,
        FileExistsError,
        FileNotFoundError,
        InterruptedError,
        IsADirectoryError,
        NotADirectoryError,
        TimeoutError,
    ]

    # Choose dead end way
    if isinstance(err, error.PaasifyError):
        err_name = err.__class__.__name__
        if isinstance(err.advice, str):
            log.warning(err.advice)

        log.error(err)
        log.critical(f"Paasify exited with: error {err.rc}: {err_name}")
        sys.exit(err.rc)

    if isinstance(err, yaml.scanner.ScannerError):
        log.critical(err)
        log.critical("Paasify exited with: YAML Scanner error (file syntax)")
        sys.exit(error.YAMLError.rc)

    if isinstance(err, yaml.composer.ComposerError):
        log.critical(err)
        log.critical("Paasify exited with: YAML Composer error (file syntax)")
        sys.exit(error.YAMLError.rc)

    if isinstance(err, yaml.parser.ParserError):
        log.critical(err)
        log.critical("Paasify exited with: YAML Parser error (file format)")
        sys.exit(error.YAMLError.rc)

    if isinstance(err, sh.ErrorReturnCode):
        log.critical(err)
        log.critical(f"Paasify exited with: failed command returned {err.exit_code}")
        sys.exit(error.ShellCommandFailed.rc)

    if isinstance(err, CaframException):
        log.critical(err)
        log.critical(f"Paasify exited with backend error. {type(err)}")
        sys.exit(error.ConfigBackendError.rc)

    if err.__class__ in oserrors:

        # Decode OS errors
        # errno = os.strerror(err.errno)
        # errint = str(err.errno)

        log.critical(f"Paasify exited with OS error: {err}")
        sys.exit(err.errno)


class CatchErrors(typer.core.TyperGroup):
    "Class to catch program exceptions and make a nice shutdown"

    def __call__(self, *args, **kwargs):
        "Used to catch program exceptions"

        try:
            return self.main(*args, **kwargs)
        except Exception as exc:
            clean_terminate(exc)

            # If we reached here, then let's go fo a good old Exception
            raise


# Main application
# ==============================

cli_app = typer.Typer(
    cls=CatchErrors,
    help=HELP_HEADER,
    rich_markup_mode="rich",
    invoke_without_command=True,
    no_args_is_help=True,
)


# Generic commands
# ==============================
help_args = {"help_option_names": ["-h", "--help"]}


@cli_app.callback(invoke_without_command=True, context_settings=help_args)
def main(
    ctx: typer.Context,
    verbose: int = typer.Option(
        0, "--verbose", "-v", count=True, min=0, max=5, help="Increase verbosity"
    ),
    working_dir: Optional[Path] = typer.Option(
        # os.getcwd(),  # For absolute paths
        # ".",          # For relative paths
        None,  # For automagic
        "-c",
        "--config",
        help="Path of paasify.yml configuration file.",
        envvar="PAASIFY_PROJECT_DIR",
    ),
    version: bool = typer.Option(
        False,
        "-V",
        "--version",
        help="Show version info",
    ),
    trace: bool = typer.Option(
        False,
        "--trace",
        help="Show traces",
    ),
):
    """

    Prepare Paasify App instance.
    """

    # pylint: disable=global-statement,invalid-name
    global log
    global PAASIFY_TRACE
    PAASIFY_TRACE = trace

    # 50: Crit
    # 40: Err
    # 30: Warn
    #   25: Notice
    # 20: Info
    #   15: Exec
    # 10: Debug
    #   5: Trace
    # 0: Not set

    # Detect extra logging
    dump_payload_log = False
    if verbose > 4:
        verbose -= 1
        dump_payload_log = True

    # Calculate log level
    verbose += 1
    verbose = 30 - (verbose * 5)
    verbose = verbose if verbose > 0 else 0

    # Logger format
    fmt = {
        0: "default",
        2: "struct",
        3: "precise",
        4: "time",
    }
    sformat = "default"
    if trace:
        sformat = fmt[2]

    # Get logger level
    log = get_logger(logger_name="paasify.cli", sformat=sformat)
    log.setLevel(level=verbose)
    # test_logging()

    # Report logging state
    if dump_payload_log:
        log.trace("Exta verbose mode enabled: data will be dumped into logs")
    else:
        log.trace("Trace mode enabled")

    if version:
        print(__version__)
        return

    app_conf = {
        "config": {
            "default_source": "default",
            "cwd": os.getcwd(),
            "root_hint": working_dir,
            "dump_payload_log": dump_payload_log,
            "no_tty": NOTTY,
            # "collections_dir": collections_dir,
        }
    }

    # Init paasify
    paasify = PaasifyApp(payload=app_conf)
    log.debug("Paasify app has been started")

    ctx.obj = {
        "paasify": paasify,
    }


# Generic commands
# ==============================


@cli_app.command("info", rich_help_panel=HELP_PROJECT_CMD)
def cli_info(
    ctx: typer.Context,
):
    """Show context infos"""
    psf = ctx.obj["paasify"]
    psf.info(autoload=True)


@cli_app.command("new", rich_help_panel=HELP_PROJECT_CMD)
def cli_new(
    ctx: typer.Context,
    source: Optional[str] = typer.Option(
        None,
        "--from",
        help="Use another project as source",
    ),
    path: Optional[str] = typer.Argument(
        ".", help="Directory of the project to create"
    ),
):
    """Create a new paasify project"""
    psf = ctx.obj["paasify"]
    psf.new_project(path, source=source)


@cli_app.command("help")
def cli_help(
    ctx: typer.Context,
):
    """Show this help message"""
    print(ctx.parent.get_help())


# Document commands
# ==============================


# pylint: disable=redefined-builtin
@cli_app.command("document_conf", rich_help_panel=HELP_DOC_CMD)
def cli_doc_conf(
    ctx: typer.Context,
    out: Optional[Path] = typer.Option(
        None,
        "-o",
        "--out",
        help="Out directory, where all files are generated",
    ),
    format_: OutputFormat = typer.Option(
        OutputFormat.json,
        "-t",
        "--format",
        help="Output format for stdout",
    ),
):
    """Build configuration schema documentation"""
    psf = ctx.obj["paasify"]
    ret = psf.cmd_config_schema(dest_dir=out)
    if not out:
        if format_ == OutputFormat.yaml:
            ret = to_yaml(ret)
        elif format_ == OutputFormat.json:
            ret = to_json(ret)
        print(ret)


@cli_app.command("document_collection", rich_help_panel=HELP_DOC_CMD)
def cli_doc_collection(
    ctx: typer.Context,
    out: Optional[Path] = typer.Option(
        None,
        "-o",
        "--out",
        help="Out directory, where all files are generated",
    ),
    mkdocs_config: Optional[Path] = typer.Option(
        None,
        "-m",
        "--mkdocs_config",
        help="Directory where to create mkdocs config, omited if empty",
    ),
    path: Optional[str] = typer.Argument(
        ".", help="Directory of the project to create"
    ),
):
    """Build collection documentation"""
    psf = ctx.obj["paasify"]
    assert isinstance(path, str), f"Got: {path}"
    psf.cmd_document(path=path, dest_dir=out, mkdocs_config=mkdocs_config)


# Does not work ????
class DocSubCmd(str, Enum):
    "Source sub commands"
    COLLECTION = "collection"
    CONF = "conf"


@cli_app.command(
    "document",
    hidden=True,
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def cli_document(
    ctx: typer.Context,
    cmd: DocSubCmd,
):
    """Command sources"""
    sub = {
        "collection": cli_doc_collection,
        "conf": cli_doc_conf,
    }
    func = sub.get(cmd)
    func(ctx, *ctx.args)


# Source commands
# ==============================


@cli_app.command("src ls", rich_help_panel=HELP_SRC_CMD)
def src_ls(
    ctx: typer.Context,
    explain: bool = typer.Option(
        False, "--explain", "-X", help="Show collection and apps"
    ),
):
    """List sources"""
    paasify = ctx.obj["paasify"]
    prj = paasify.load_project()
    prj.sources.cmd_ls(explain=explain)


@cli_app.command("src install", rich_help_panel=HELP_SRC_CMD)
def src_install(
    ctx: typer.Context,
):
    """Install sources"""
    paasify = ctx.obj["paasify"]
    prj = paasify.load_project()
    prj.sources.cmd_install()


@cli_app.command("src update", rich_help_panel=HELP_SRC_CMD)
def src_update(
    ctx: typer.Context,
):
    """Update sources"""
    paasify = ctx.obj["paasify"]
    prj = paasify.load_project()
    prj.sources.cmd_update()


@cli_app.command("src tree", rich_help_panel=HELP_SRC_CMD)
def src_tree(
    ctx: typer.Context,
):
    """Show source tree"""
    paasify = ctx.obj["paasify"]
    prj = paasify.load_project()
    prj.sources.cmd_tree()


class SrcSubCmd(str, Enum):
    "Source sub commands"

    LS = "ls"
    UPDATE = "update"
    TREE = "tree"
    INSTALL = "install"


@cli_app.command(
    "src",
    hidden=True,
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def cli_src(
    ctx: typer.Context,
    cmd: SrcSubCmd,
):
    """Command sources"""
    sub = {
        "ls": src_ls,
        "update": src_update,
        "tree": src_tree,
        "install": src_install,
    }
    func = sub.get(cmd)
    return ctx.invoke(func, ctx)


# Stack commands (Base)
# ==============================


@cli_app.command("ls", rich_help_panel=HELP_STACKS_CMD)
def cli_ls(
    ctx: typer.Context,
    explain: bool = typer.Option(
        False, "--explain", "-X", help="Show stacks structure"
    ),
    stack: Optional[str] = typer.Argument(
        None,
        help="Stack to target, current directory or all, only when explain is enabled",
    ),
):
    """List all stacks"""
    paasify = ctx.obj["paasify"]
    prj = paasify.load_project()
    if explain:
        prj.stacks.cmd_stack_explain(stack_names=stack)
    else:
        prj.stacks.cmd_stack_ls()


@cli_app.command("build", rich_help_panel=HELP_STACKS_CMD)
def cli_assemble(
    ctx: typer.Context,
    explain: bool = typer.Option(
        False, "--explain", "-X", help="Show diff between modifications"
    ),
    stack: Optional[str] = typer.Argument(
        None,
        help="Stack to target, current directory or all",
    ),
):
    """Build docker-files"""

    paasify = ctx.obj["paasify"]
    prj = paasify.load_project()
    prj.stacks.cmd_stack_assemble(stack_names=stack, explain=explain)


@cli_app.command("up", rich_help_panel=HELP_STACKS_CMD)
def cli_up(
    ctx: typer.Context,
    logs: bool = typer.Option(
        False, "--logs", "-l", help="Show running logs after action"
    ),
    stack: Optional[str] = typer.Argument(
        None,
        help="Stack to target, current directory or all",
    ),
):
    """Start docker stack"""
    paasify = ctx.obj["paasify"]
    prj = paasify.load_project()
    prj.stacks.cmd_stack_up(stack_names=stack)

    if logs:
        prj.stacks.cmd_stack_logs(stack_names=stack, follow=True)


@cli_app.command("down", rich_help_panel=HELP_STACKS_CMD)
def cli_down(
    ctx: typer.Context,
    stack: Optional[str] = typer.Argument(
        None,
        help="Stack to target, current directory or all",
    ),
):
    """Stop docker stack"""
    paasify = ctx.obj["paasify"]
    prj = paasify.load_project()
    prj.stacks.cmd_stack_down(stack_names=stack)


@cli_app.command("ps", rich_help_panel=HELP_STACKS_CMD)
def cli_ps(
    ctx: typer.Context,
    stack: Optional[str] = typer.Argument(
        None,
        help="Stack to target, current directory or all",
    ),
):
    """Show docker stack instances"""
    paasify = ctx.obj["paasify"]
    prj = paasify.load_project()
    prj.stacks.cmd_stack_ps(stack_names=stack)


@cli_app.command("logs", rich_help_panel=HELP_STACKS_CMD)
def cli_logs(
    ctx: typer.Context,
    follow: bool = typer.Option(False, "--follow", "-f"),
    stack: Optional[str] = typer.Argument(
        None,
        help="Stack to target, current directory or all",
    ),
):
    """Show stack logs"""
    paasify = ctx.obj["paasify"]
    prj = paasify.load_project()
    prj.stacks.cmd_stack_logs(stack_names=stack, follow=follow)


@cli_app.command("vars", rich_help_panel=HELP_STACKS_CMD)
def cli_vars(
    ctx: typer.Context,
    vars: Optional[str] = typer.Option(
        None,
        help="List of vars comma separated to show only",
    ),
    explain: bool = typer.Option(
        False, "--explain", "-X", help="Show running logs after action"
    ),
    stack: Optional[str] = typer.Argument(
        None,
        help="Stack to target, current directory or all",
    ),
):
    """Dump stack variables"""
    paasify = ctx.obj["paasify"]
    prj = paasify.load_project()
    vars = vars.split(",") if vars else True
    prj.stacks.cmd_stack_vars(stack_names=stack, vars_=vars, explain=explain)


# Stack commands (Helpers)
# ==============================


@cli_app.command("apply", rich_help_panel=HELP_STACKS_HELPERS)
def cli_apply(
    ctx: typer.Context,
    logs: bool = typer.Option(
        False, "--logs", "-l", help="Show running logs after action"
    ),
    stack: Optional[str] = typer.Argument(
        None,
        help="Stack to target, current directory or all",
    ),
):
    """Build and apply stack"""
    paasify = ctx.obj["paasify"]
    prj = paasify.load_project()
    prj.stacks.cmd_stack_apply(stack_names=stack)

    if logs:
        prj.stacks.cmd_stack_logs(stack_names=stack, follow=True)


@cli_app.command("recreate", rich_help_panel=HELP_STACKS_HELPERS)
def cli_recreate(
    ctx: typer.Context,
    logs: bool = typer.Option(
        False, "--logs", "-l", help="Show running logs after action"
    ),
    stack: Optional[str] = typer.Argument(
        None,
        help="Stack to target, current directory or all",
    ),
):
    """Stop, rebuild and create stack"""
    paasify = ctx.obj["paasify"]
    prj = paasify.load_project()
    prj.stacks.cmd_stack_recreate(stack_names=stack)

    if logs:
        prj.stacks.cmd_stack_logs(stack_names=stack, follow=True)


# Top levels helpers
# ==============================


def app():
    "Return a Paasify App instance"

    try:
        return cli_app()

    # pylint: disable=broad-except
    except Exception as err:

        # Developper catchall
        log.error(traceback.format_exc())
        log.critical(f"Uncatched error {err.__class__}; this may be a bug!")
        log.critical("Exit 1 with bugs")
        sys.exit(1)


if __name__ == "__main__":
    app()



# -*- coding: utf-8 -*-
"""Paasify Engine management

This class helps to deal with docker engine versions
"""

# pylint: disable=logging-fstring-interpolation
# pylint: disable=invalid-name

import os
import re
import logging
import json

from pprint import pprint  # noqa: F401

import semver
import sh

from cafram.utils import _exec
from cafram.nodes import NodeMap

import paasify.errors as error
from paasify.common import cast_docker_compose
from paasify.framework import PaasifyObj


log = logging.getLogger(__name__)


def bin2utf8(obj):
    "Transform sh output bin to utf-8"

    if hasattr(obj, "stdout"):
        obj.txtout = obj.stdout.decode("utf-8").rstrip("\n")
    else:
        obj.txtout = None
    if hasattr(obj, "stderr"):
        obj.txterr = obj.stderr.decode("utf-8").rstrip("\n")
    else:
        obj.txterr = None
    return obj


#####################

# https://www.docker.com/blog/announcing-compose-v2-general-availability/
# v1 vs v2
#  v2 fully support --project-name while v1 does not


class EngineCompose(NodeMap, PaasifyObj):
    "Generic docker-engine compose API"

    _node_parent_kind = ["Stack"]

    version = None
    docker_file_exists = False
    # docker_file_path = None
    arg_prefix = []

    conf_default = {
        "stack_name": None,
        "stack_path": None,
        "docker_file": "docker-compose.yml",
        # "docker_file_path": None,
    }

    ident = "default"

    compose_bin = "docker"
    # compose_pre = [
    #        "compose",
    #        "--file", "myfile.yml",
    #        "--project-name", "project-name",
    #        ]

    # compose_bin = "docker-compose"
    # compose_pre = [
    #        "--file", "myfile.yml",
    #        "--project-name", "project-name",
    #        ]

    def node_hook_init(self):
        "Create instance attributes"

        self.docker_file_path = None
        self.arg_prefix_full = []
        self.arg_prefix = []

    def node_hook_children(self):
        "Create stack context on start"

        # Get parents
        # stack = self._node_parent
        # prj = stack.prj

        # Init object
        # self.stack_name = self.stack_name
        # self.stack_path = self.stack_path
        # self.docker_file_path = self.docker_file_path or os.path.join(self.stack_path, self.docker_file)
        # pprint (self.__dict__)
        self.docker_file_path = os.path.join(self.stack_path, self.docker_file)
        # pprint (self.__dict__)

        # dsfsdf

        # Pre build args
        self.arg_prefix = [
            "compose",
            "--project-name",
            f"{self.stack_name}",
            # "--project-directory", f"{self.stack_path}",
        ]
        self.arg_prefix_full = [
            "compose",
            "--project-name",
            f"{self.stack_name}",
            "--file",
            f"{self.docker_file_path}",
        ]

    def run(self, cli_args=None, command=None, logger=None, **kwargs):
        "Wrapper to execute commands"

        command = command or self.compose_bin
        cli_args = cli_args or []

        self.log.exec(f"Run: {command} {' '.join(cli_args)}")
        # result = _exec(command, cli_args=cli_args, logger=self.log, **kwargs)
        result = _exec(command, cli_args=cli_args, **kwargs)

        if result:
            result = bin2utf8(result)
            log.notice(result.txtout)

        return result

    def require_stack(self):
        "Ensure stack context"

        if not self.stack_name:
            assert False, "Command not available for stacks!"

    def require_compose_file(self):
        "Raise an exception when compose file is absent"

        self.require_stack()

        if not os.path.isfile(self.docker_file_path):
            self.log.warning("Please build stack first")
            raise error.BuildStackFirstError("Docker file is not built yet !")

    def assemble(self, compose_files, env_file=None, env=None):
        "Generate docker-compose file"

        self.require_stack()

        cli_args = list(self.arg_prefix)

        if env_file:
            cli_args.extend(["--env-file", env_file])
        for file in compose_files:
            cli_args.extend(["--file", file])
        cli_args.extend(
            [
                "config",
                # "--no-interpolate",
                # "--no-normalize",
            ]
        )

        env_string = env or {}
        env_string = {
            k: cast_docker_compose(v)
            for k, v in env.items()
            if isinstance(
                v,
                (
                    str,
                    int,
                    bool,
                    #  type(None) # Do we want None as well ?
                ),
            )
        }

        out = self.run(cli_args=cli_args, _out=None, _env=env_string)
        return out

    # pylint: disable=invalid-name
    def up(self, **kwargs):
        "Start containers"

        self.require_compose_file()
        cli_args = self.arg_prefix_full + [
            "up",
            "--detach",
        ]
        out = self.run(cli_args=cli_args, **kwargs)
        return out

    def down(self, **kwargs):
        "Stop containers"

        self.require_stack()
        # cli_args = list(self.arg_prefix)
        cli_args = self.arg_prefix_full + [
            # "--project-name",
            # self.stack_name,
            "down",
            "--remove-orphans",
        ]

        try:
            out = self.run(cli_args=cli_args, **kwargs)
            # out = _exec("docker-compose", cli_args, **kwargs)
            # if out:
            #    bin2utf8(out)
            #    log.notice(out.txtout)

        # pylint: disable=no-member
        except sh.ErrorReturnCode_1 as err:
            bin2utf8(err)

            # This is U.G.L.Y
            if "has active endpoints" not in err.txterr:
                raise error.DockerCommandFailed(f"{err.txterr}")

        return out

    def logs(self, follow=False):
        "Return container logs"

        self.require_stack()
        sh_options = {}
        cli_args = self.arg_prefix + [
            # "--project-name",
            # self.stack_name,
            "logs",
        ]
        if follow:
            cli_args.append("-f")
            sh_options["_fg"] = True

        out = self.run(cli_args=cli_args, **sh_options)
        print(out)

    # pylint: disable=invalid-name
    def ps(self):
        "Return container processes"

        self.require_stack()

        # Prepare command
        cli_args = self.arg_prefix + [
            # "compose",
            # "--project-name",
            # self.stack_name,
            "ps",
            "--all",
            "--format",
            "json",
        ]
        result = self.run(cli_args=cli_args, _out=None)

        # Report output from json
        stdout = result.txtout
        payload = json.loads(stdout)
        for svc in payload:

            # Get and filter interesting ports
            published = svc["Publishers"] or []
            published = [x for x in published if x.get("PublishedPort") > 0]

            # Reduce duplicates
            for pub in published:
                if pub.get("URL") == "0.0.0.0":
                    pub["URL"] = "::"

            # Format port strings
            exposed = []
            for port in published:
                src_ip = port["URL"]
                src_port = port["PublishedPort"]
                dst_port = port["TargetPort"]
                prot = port["Protocol"]

                r = f"{src_ip}:{src_port}->{dst_port}/{prot}"
                exposed.append(r)

            # Remove duplicates ports and show
            exposed = list(set(exposed))
            print(
                f"  {svc['Project'] :<32} {svc['ID'][:12] :<12} {svc['Name'] :<40} {svc['Service'] :<16} {svc['State'] :<10} {', '.join(exposed)}"
            )


class EngineComposeV2(EngineCompose):
    "Docker-engine: Support for version until 2.6"

    ident = "docker compose 2"


class EngineComposeV1(EngineCompose):
    "Docker-engine: Support for version until 1.29"

    ident = "docker-compose 1"

    # pylint: disable=invalid-name
    def ps(self):
        cli_args = [
            "--file",
            self.docker_file_path,
            "ps",
            "--all",
        ]

        result = _exec("docker-compose", cli_args, _fg=True)

        return result


class EngineCompose_16(EngineCompose):
    "Docker-engine: Support for version until 1.6"

    ident = "docker-compose-1.6"


class EngineDetect:
    "Class helper to retrieve the appropriate docker-engine class"

    versions = {
        "docker": {
            "20.10.17": {},
        },
        "docker-compose": {
            "2.0.0": EngineComposeV2,
            "1.0.0": EngineComposeV1,
            # "2.6.1": EngineCompose_26,
            # "1.29.0": EngineCompose_129,
            # "1.6.3": EngineCompose_16,
        },
        "podman-compose": {},
    }

    def detect_docker_compose(self):
        "Detect current version of docker compose. Return a docker-engine class."

        # pylint: disable=no-member

        # Try docker-compose v1

        # Try docker compose v2
        out = "No output for command"
        patt = r"version v?(?P<version>(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<patch>[0-9]+))"
        try:
            log.notice("This can take age when debugger is enabled...")
            out = _exec("docker", ["compose", "version"])
            # TOFIX: This takes ages in debugger, when above _log_msg is unset ?
            # out = cmd('--version')
            bin2utf8(out)
        except sh.ErrorReturnCode as err:
            # raise error.DockerUnsupportedVersion(
            #    f"Impossible to guess docker-compose version: {out}"
            # ) from err

            # pylint: disable=no-member
            try:

                # cmd = sh.Command("docker-compose", _log_msg='paasify')
                log.notice("This can take age when debugger is enabled...")
                out = _exec("docker-compose", ["--version"])
                # TOFIX: This takes ages in debugger, when above _log_msg is unset ?
                # out = cmd('--version')
                bin2utf8(out)
            except sh.ErrorReturnCode:
                raise error.DockerUnsupportedVersion(
                    f"Impossible to guess docker-compose version: {out}"
                ) from err

        # Scan version
        match = re.search(patt, out.txtout)
        if match:
            version = match.groupdict()
        else:
            msg = f"Output format of docker-compose is not recognised: {out.txtout}"
            raise error.DockerUnsupportedVersion(msg)
        curr_ver = semver.VersionInfo.parse(version["version"])

        # Scan available versions
        versions = sorted(
            list(map(semver.VersionInfo.parse, self.versions["docker-compose"].keys()))
        )
        versions.reverse()
        match = None
        for version in versions:
            works = curr_ver.match(f">={version}")
            if works:
                match = str(version)
                break

        if not match:
            raise error.DockerUnsupportedVersion(
                f"Version of docker-compose is not supported: {curr_ver}"
            )

        cls = self.versions["docker-compose"][match]
        cls.version = match
        cls.name = "docker-compose"
        cls.ident = match
        return cls

    def detect(self, engine=None):
        "Return the Engine class that match engine string"

        if not engine:
            log.info("Guessing best docker engine ...")
            obj = self.detect_docker_compose()
        else:

            if engine not in self.versions["docker-compose"]:
                versions = list(self.versions["docker-compose"].keys())
                log.warning(f"Please select engine one of: {versions}")
                raise error.DockerUnsupportedVersion(
                    f"Unknown docker-engine version: {engine}"
                )
            obj = self.versions["docker-compose"][engine]
        # if not result:
        #     raise error.DockerUnsupportedVersion(f"Can;t find docker-compose")

        log.debug(f"Detected docker-compose version: {obj.version}")

        return obj



"""
Config Node classes
"""
# pylint: disable=too-many-lines
# pylint: disable=too-few-public-methods
# pylint: disable=arguments-renamed
# pylint: disable=arguments-differ
# pylint: disable=unused-argument
# pylint: disable=logging-fstring-interpolation

import os
import copy
import inspect
import textwrap
import json
import logging

# from pprint import pprint

import jsonschema

from cafram.utils import serialize, json_validate, truncate

from cafram.base import (
    Base,
    DictExpected,
    ListExpected,
    NotExpectedType,
    ClassExpected,
    InvalidSyntax,
    SchemaError,
    ApplicationError,
)


_log = logging.getLogger(__name__)

# Display complete payload in logs
TRUNCATE = -1

# Functions
# =====================================


def map_container_class(payload):  # map_container_class
    "Map list or dict to cafram classes, otherwise keep value as is"

    # if payload is None:
    #     return NodeVal
    if isinstance(payload, dict):
        return NodeMap
        # return NodeDict

    if isinstance(payload, list):
        return NodeList

    return type(payload)


def map_node_class(payload):  # map_node_class
    "Map anything to cafram classes"

    if isinstance(payload, dict):
        return NodeMap
        # return NodeDict
    if isinstance(payload, list):
        return NodeList

    return NodeVal


# Simple attributes class
# =====================================

# pylint: disable=too-many-instance-attributes
class NodeVal(Base):
    """
    Base configuration class

    * conf_default: Holds the default configuration.
    if it's a dict, it will be merged with acutal user config, otherwise

    * conf_struct: If a value si present in payload, an instance will be created with
    default data, unless the key is not present (in case of dicts)
    """

    # Class public parameters
    conf_default = None
    conf_schema = None
    conf_ident = None
    conf_children = None

    # Class internal attributes
    _nodes = None
    _node_root = None
    _node_parent = None
    _node_parent_kind = []

    _node_conf_raw = None
    _node_conf_parsed = None
    _node_autoconf = 0
    _node_lvl = 0
    _node_kind = "Value"

    def __init__(self, *args, parent=None, payload=None, autoconf=None, **kwargs):

        # Call parents
        # pylint: disable=super-with-arguments
        super(NodeVal, self).__init__(*args, **kwargs)

        # Register parent
        self._node_parent = parent or self
        assert isinstance(
            self._node_parent, NodeVal
        ), f"Parent of {self} is not a NodeVal descendant object, got: {self._node_parent}"
        self._node_root = getattr(parent, "_node_root", None) or self

        # Enforce parrent type
        if self._node_parent_kind:
            cls_name = self._node_parent.__class__.__name__
            assert (
                cls_name in self._node_parent_kind
            ), f"Bug: Parent node '{self._node_parent}' allowed to create {self}, only: {self._node_parent_kind}"

        # Manage autoconf levels
        if autoconf is None:
            autoconf = getattr(self._node_parent, "_node_autoconf", 0)
        self._node_autoconf = (autoconf - 1) if autoconf > 0 else autoconf

        # Manage node level
        self._node_lvl = getattr(self._node_parent, "_node_lvl", -1) + 1

        # Manage log indentation
        class CustomAdapter(logging.LoggerAdapter):
            "LoggerADpater to manage logging indentation"

            def process(self, msg, kwargs):
                indent = kwargs.pop("indent", self.extra["indent"])
                return f"{indent}{msg}", kwargs

            def __getattr__(self, name):
                "Forward unknown attributes such as custom levels in logger"
                return getattr(self.logger, name)

        indent = "| " * self._node_lvl  # +  "  "
        self._node_log = CustomAdapter(logging.getLogger('cafram'), {"indent": indent})

        # Auto init object
        self.node_hook_init()
        self.deserialize(payload)

    # Serialization
    # -----------------

    def deserialize(self, payload):
        "Transform json to object"

        # 0. Init
        # -------------------

        # pylint: disable=W0212
        self._nodes = self.__class__._nodes
        self._node_conf_raw = payload

        self._node_log.info(f"> Deserialize Node: {self}")

        # 1. Parse config
        # -------------------
        # 1.1 Validate config against schema
        # 1.2 Run transform hook (node_hook_transform)
        # 1.3 Apply defaults from conf_children or conf_default
        # 1.4 User report

        self._node_log.debug(f"  1. Validate {self} config")
        payload1 = self._node_conf_validate(payload)
        self._node_log.debug(f"  2. Hook: node_hook_transform {self} config")
        payload2 = self.node_hook_transform(payload1)
        self._node_log.debug(f"  3. Apply conf defaults {self} config")
        payload3 = self._node_conf_defaults(payload2)

        if payload1 != payload3:
            self._node_log.debug(f"    3.3 Payload transformation for: {self}")
            self._node_log.debug(f"         in: {payload1}")
            self._node_log.debug(f"        out: {payload3}")

        # 2. Apply config
        # -------------------
        # 2.1 Inject onfig into node
        # 2.2 Run conf hook (node_hook_conf)

        self._node_conf_parsed = payload3
        self._node_log.debug(f"  4. Hook: node_hook_conf {self} config")
        self.node_hook_conf()

        # 3. Create children
        # -------------------
        # 3.1 Create children nodes
        # 3.2 Run children hook (node_hook_children)
        # 3.3 Preset default ident

        self._node_log.debug(f"  5. Build {self} config")
        self._node_conf_build()
        self._node_log.debug(f"  6. Hook: node_hook_children {self} config")
        self.node_hook_children()
        if self.conf_ident:
            try:
                self.ident = self.conf_ident.format(**locals())
            except AttributeError as err:
                msg = f"Bug: on '{self}.conf_ident={self.conf_ident}', {err.args[0]}"
                raise ApplicationError(msg) from err

        self._node_log.debug(f"  7. Node {self} has been created")

        self.node_hook_final()

    def serialize(self, mode="parsed"):
        "Transform object to json"

        if mode == "raw":
            value = self._node_conf_raw
        elif mode == "parsed":
            value = self._node_conf_parsed
        elif mode == "default":
            value = self.conf_default
        else:
            raise Exception(f"Unknown mode: {mode}")

        return value

    # User hooks
    # -----------------

    # Available hooks:
    # node_hook_transform:
    #   - Payload modifications only
    # node_hook_conf
    #   - Payload is done, preset values
    # node_hook_children
    #   - Once the children has been created

    def node_hook_init(self):
        "Placeholder to executes in __init__ object"

    def node_hook_transform(self, payload):
        "Placeholder to transform config after validation"
        return payload

    def node_hook_conf(self):
        "Placeholder to executes after configuration build"

    def node_hook_children(self):
        "Placeholder to transform object once onfig has been done"

    def node_hook_final(self):
        "Placeholder to transform object once everything done"

    # Configuration parser
    # -----------------

    def _node_conf_defaults(self, payload):
        """Return payload or default value"""

        return payload or self.conf_default

    def _node_conf_validate(self, payload):
        """Validate config against schema

        This function will only works if:
        * self.conf_schema is not None
        * self.conf_schema conatins the "$schema" key
        """

        # old_payload = payload

        # pylint: disable=E1135
        if isinstance(self.conf_schema, dict):
            if "$schema" in self.conf_schema:

                try:
                    self._node_log.info("    1.2 Validate payload")
                    payload = json_validate(self.conf_schema, payload)
                except jsonschema.exceptions.ValidationError as err:

                    self._node_log.critical(f"Value: {err.instance}")
                    self._node_log.critical(f"Payload: {payload}")
                    raise SchemaError(
                        f"Schema validation error for {self}: {err.message}"
                    ) from err

                except jsonschema.exceptions.SchemaError as err:
                    # print("Bug in schema for ", self)
                    # print(err)

                    # print("PAYLOAD")
                    # pprint(payload)
                    # print("SCHEMA")
                    # pprint(self.conf_schema)
                    # print("BBBUUUUUGGGGGGG on schema !!!")
                    # # print(traceback.format_exc())
                    raise Exception(err) from err

                # if old_payload != payload:
                #     print("OLD CONFIG WITHOUT DEFAULTS", old_payload)
                #     print("NEW CONFIG WITH DEFAULTS", payload)

        # else:
        #     print(f"NO SCHEMA VALIDATION FOR {self}")

        return payload

    # Simple Class implementation
    # -----------------
    def _node_conf_build(self):
        "Just assign the value, thats all"
        self._nodes = None

    # Misc
    # -----------------

    def from_json(self, payload):
        "Load from json string"

        payload = json.loads(payload)
        return self.deserialize(payload)

    # Node management
    # -----------------

    # Methods:
    # get_children
    #   - return only node objects
    #   - support recur
    # get_value
    #   - return ALL but node objects

    def get_children(self, lvl=0, explain=False, leaves=False):
        """A nodeVal can't have a children, so always return None"""
        result = None
        if leaves:
            result = self
        if explain:
            return {
                "obj": self,
                "value": self.get_value(),
                f"subtype (auto:{self._node_autoconf})": self.conf_children,
                "children": result,
            }
        return result

    def get_value(self, **kwargs):
        """Return the _nodes value (value+children)"""

        return self._node_conf_parsed

    # Sibling management
    # -----------------

    def is_root(self):
        """Return True if object is root"""
        if self._node_parent and self._node_parent == self:
            return True
        return False

    def get_parent(self):
        """Return first parent"""
        return self._node_parent or None

    def get_parent_root(self):
        """Return root parent object"""
        return self._node_root

    def get_parents(self):
        "Return all parent of the object"

        parents = []
        current = self
        parent = self._node_parent or None
        while parent is not None and parent != current:
            if parent not in parents:
                parents.append(parent)
                current = parent
                parent = getattr(current, "_node_parent")

        return parents

    # Dumper
    # -----------------

    # pylint: disable=redefined-builtin
    def dump(self, explain=True, all=True, **kwargs):
        """Output a dump of the object, helpful for troubleshooting prupose"""

        # pylint: disable=super-with-arguments
        super(NodeVal, self).dump(**kwargs)

        _node_conf_parsed = self._node_conf_parsed
        _node_conf_raw = self._node_conf_raw

        print("  Node info:")
        print("  -----------------")
        print(f"    Node level: {self._node_lvl}")
        print(f"    Node root: {self._node_root}")
        print(f"    Node parents: {self.get_parents()}")
        print("")

        if all and _node_conf_parsed is not None:
            msg = "(same as Raw Config)"
            if _node_conf_raw != _node_conf_parsed:
                msg = "(different from Raw Config)"

                print("  Raw Config:")
                print("  -----------------")
                out = serialize(_node_conf_raw, fmt="yaml")
                print(textwrap.indent(out, "    "))

            print("  Parsed Config:", msg)
            print("  -----------------")
            out = serialize(_node_conf_parsed, fmt="yaml")
            print(textwrap.indent(out, "    "))
            # print ("")

        if all:
            print("  Children: (All)")
            print("  -----------------")

            children = self.get_children(lvl=-1, explain=False, leaves=True)
            out = serialize(children)
            # out = pformat(children, indent=2, width=5)
            print(textwrap.indent(out, "    "))
            print("")

            # children_all = self.get_children(lvl=-1, explain=False, leaves=True)
            # if children != children_all:
            #     out = serialize(children_all)
            #     print(textwrap.indent(out, "    "))
            #     print("")

            print("  Value:")
            print("  -----------------")
            children = self.get_value(explain=explain)
            # out = serialize(children, fmt="yaml")
            out = serialize(children, fmt="json")
            print(textwrap.indent(out, "    "))
            print("")

        print("  Whole config:")
        print("  -----------------")
        children = self.get_value(lvl=-1)
        # out = serialize(children, fmt="yaml")
        out = serialize(children, fmt="json")
        print(textwrap.indent(out, "    "))
        print("")

        # out = pformat(children)
        # print (textwrap.indent(out, '    '))

        # print("\n")

    def show_childs(self, lvl=0):
        "Display a nice tree view of all objects"

        lvl += 1
        indent = "| " * lvl
        children = self.get_children()

        # pylint: disable=line-too-long
        if isinstance(children, dict):
            for name, child in children.items():
                head = truncate(f"{indent}{name}:", 38)
                val = truncate(f"{child.get_value()}")
                print(
                    f"{head:<40}{child._node_kind:<10}{id(child)}:{child.__class__.__name__:<30}{val}"
                )
                child.show_childs(lvl)
        elif isinstance(children, list):
            count = -1
            for child in children:
                count += 1
                head = truncate(f"{indent} - <item_{count}>:", 38)
                val = truncate(f"{child.get_value()}")
                print(
                    f"{head:<40}{child._node_kind:<10}{id(child)}:{child.__class__.__name__:<30}{val}"
                )
                child.show_childs(lvl)
        else:
            assert False, f"Damn: {self} => {children}"


# Test Class Data
# =====================================


class NodeList(NodeVal):
    """NodeList"""

    _nodes = []
    _node_conf_parsed = []
    _node_kind = "List"

    # Overrides
    # -------------------

    def _node_conf_defaults(self, payload):
        """Return payload or default list (NodeList)"""

        payload = payload or self.conf_default or []
        if not isinstance(payload, list):
            raise ListExpected(f"A list was expected for {self}, got: {payload}")

        return payload

    def _node_conf_build(self):
        "Just assign the value, thats all NodeList"

        payload = self._node_conf_parsed
        if not payload:
            self._nodes = []
            return

        cls = self.conf_children

        results = []
        count = -1
        are_children = False
        for item in payload:
            count += 1
            ident = f"{self.ident}_{count}"

            if self._node_autoconf != 0:
                # If not found, first element of the list determine class
                # cls = cls or map_node_class(item)
                cls = cls or map_container_class(item)
                # print ("AUTOGUESS CLASS: ", cls)

            if cls:
                if not inspect.isclass(cls):
                    raise ClassExpected(
                        f"A class was expected for {self}.conf_struct, got {type(cls)}: {cls}"
                    )

                if issubclass(cls, NodeVal):
                    are_children = True
                    result = cls(parent=self, ident=ident, payload=item)
                elif cls:

                    are_children = False

                    if item and not isinstance(item, cls):
                        msg = f"""Wrong type in list: '{item}' is not
                        a '{cls.__name__}' type in '{payload}'"""
                        raise NotExpectedType(msg)

                    result = item
                    if item:
                        result = cls(item)

                    # try:

                    # except (ValueError, TypeError) as err:
                    #     msg = f"""Wrong type in list: '{item}' is not a
                    # '{cls.__name__}' type in '{payload}'"""
                    #     raise NotExpectedType(msg) from err

                results.append(result)
            else:
                results.append(item)


        if are_children:
            self._nodes = results
        else:
            self._node_conf_parsed = results

    def __iter__(self):
        return self._nodes.__iter__()

    def get_children(self, lvl=0, explain=False, leaves=False):
        "Return NodeList childs"
        result = []
        for child in self._nodes:
            if lvl != 0:
                value = child.get_children(lvl=lvl - 1, explain=explain, leaves=leaves)
            else:
                value = child

            if leaves:
                value = value or child

            result.append(value)

        if explain:
            return {
                "obj": self,
                "value": self.get_value(),
                f"subtype (auto:{self._node_autoconf})": self.conf_children,
                "children": result,
            }

        return result

    def get_value(self, lvl=0, explain=False):
        "Return NodeList value"
        result = []
        for child in self._nodes:
            if lvl != 0:
                result.append(child.get_value(lvl=lvl - 1, explain=explain))
            else:
                result.append(child)

        return result or self._node_conf_parsed


# NodeDict
# =====================================


class NodeDictItem:
    """Children configuration for NodeDict"""

    def __init__(
        self,
        *args,
        key=None,
        cls=None,
        action="set",
        hook=None,
        default=None,
        attr="__UNSET__",
        **kwargs,
    ):

        self.key = key
        self.attr = key if attr == "__UNSET__" else attr

        self._ident = None
        self.hook = hook

        self.cls = cls or None
        self.default = default or None
        self.action = action

    @property
    def ident(self) -> str:
        "Return ident"
        return self.attr or self.key

    def __repr__(self):
        result = [f"{key}={val}" for key, val in self.__dict__.items() if val]
        result = "|".join(result)
        return f"Remap:{result}"


class NodeDictItemManager:
    "Manage DictItemChildren"

    def __init__(self, conf_children, payload=None, autoconf=None, log=None):
        self._node_log = log or _log

        self.data = self.load_conf(conf_children, payload=payload, autoconf=autoconf)

    def __iter__(self):
        "Forward iteration"
        return self.data.__iter__()

    def load_conf(self, conf_children, payload=None, autoconf=0):
        "Load conf and generate internal NodeDictItem"

        # Conf_children default behavior: auto from dict
        # conf_children = self.conf_children or {}
        payload = payload or {}
        log_prefix = "    3."

        # 1. Leave as is
        if isinstance(conf_children, list):
            log_mode = "Keep list as is"

        # 2. Direct generate
        elif inspect.isclass(conf_children):
            log_mode = f"Create childs from class: {conf_children}"
            conf_children = [
                {"key": key, "default": val, "cls": conf_children}
                for key, val in payload.items()
            ]

        # 3. Auto generate
        elif autoconf != 0:
            log_mode = "Automap sub containers"
            conf_children = [
                {"key": key, "default": val, "cls": map_container_class(val)}
                for key, val in payload.items()
            ]

        # 4. Auto guess from payload
        elif not conf_children:
            log_mode = "Guess from payload types"
            conf_children = [
                {
                    "key": key,
                    "default": val,
                    "cls": type(val) if val is not None else None,
                }
                for key, val in payload.items()
            ]

        # Actually build conf_Struct
        _payload = str(payload)
        self._node_log.info(
            log_prefix
            + f"1 Children config: {log_mode} with {truncate(_payload, max=TRUNCATE)}"
        )
        conf_struct = [NodeDictItem(**conf) for conf in conf_children]
        self._node_log.debug(
            log_prefix + f"1 Children plan: {truncate(conf_struct, max=TRUNCATE)}"
        )

        # Developper sanity check
        if not isinstance(conf_struct, list):
            raise ListExpected(
                f"A list was expected for conf_children, got : {conf_struct}, {conf_children}"
            )

        return conf_struct

    def clean(self, payload):
        "Clean payload"

        for item_def in self.data:

            key = item_def.key
            attr = item_def.attr
            action = item_def.action

            # Get value
            if key is None:
                continue

            # Check value
            value = payload.get(key, item_def.default or None)
            if action == "unset":
                value = value or None
            elif action == "drop":
                if key in payload:
                    del payload[key]
                if attr in payload:
                    del payload[attr]
                continue

            # Remove old key
            payload[key] = value
            if key != attr and attr in payload:
                del payload[attr]

        return payload

    def build_it(self, node):
        "Actually create children nodes/values"

        assert isinstance(node, NodeDict), f"BUG: Wrong type, expected NodeDict: {self}"
        log_prefix = "    5."
        self._node_log.debug(log_prefix + f"1 Build {node} children ...")

        # 2. Process each children
        for item_def in self.data:

            key = item_def.key
            attr = item_def.attr
            cls = item_def.cls
            action = item_def.action
            hook = item_def.hook

            # Get value
            value = None
            if key:
                value = node.get_value().get(key)

            # Check action
            if not value:
                if action == "unset":
                    # value = None
                    cls = None
                elif action == "drop":
                    continue

            log_msg = None
            if cls:
                # pylint: disable=line-too-long
                # Instanciate or cast value
                if issubclass(cls, NodeVal):

                    self._node_log.info(
                        log_prefix
                        + f"2 Instanciate Children Node object: {attr}={cls}({truncate(value, max=TRUNCATE)})"
                    )
                    self._node_log.info(" ")
                    child = cls(parent=node, ident=item_def.ident, payload=value)

                    log_msg = f"Instanciated Children Node object: {attr}={cls}({truncate(value, max=TRUNCATE)})"
                    # Update parsed conf
                    if attr:
                        node.add_child(attr, child)
                    value = child.serialize(mode="parsed")

                else:
                    if not value:
                        log_msg = f"Instanciate value: empty object: {attr}={cls}()"
                        value = cls()
                    elif isinstance(value, cls):
                        log_msg = f"Instanciate value: cast value: {attr}={cls}({truncate(value, max=TRUNCATE)})"
                        value = cls(value)
                    else:
                        log_msg = f"Instanciate value: raw class: {attr}={cls}()"
                        try:
                            value = cls(value)
                        except Exception as err:
                            self._node_log.critical(
                                f"{log_msg}\nType mismatch between for {self}: {cls} and {truncate(value, max=TRUNCATE)}."
                            )
                            assert False, f"Set correctly the exception: {err.__class__}"
                            raise NotExpectedType(err) from err

            else:
                # Forward value
                log_msg = f"Instanciate direct assignment: {attr}={value}"
                # value = value

            if log_msg:
                self._node_log.info(log_prefix + "3 " + log_msg)

            # Patch original configuration
            # pylint: disable=protected-access
            if attr:
                node._node_conf_parsed[attr] = value
            if key and attr != key:
                del node._node_conf_parsed[key]

            if hook:
                fun = getattr(node, hook)
                self._node_log.debug(log_prefix + f"3 Execute hook: {hook}, {fun}")
                fun()


class NodeDict(NodeVal):
    "Node Dict container"

    _nodes = {}
    _node_conf_struct = None
    _node_conf_parsed = {}
    _node_kind = "Dict"

    # TODO: https://www.pythonlikeyoumeanit.com/Module4_OOP/Special_Methods.html
    # __len__
    # __getitem__(self, key)
    # __setitem__(self, key, item)
    # __contains__(self, item)
    # __iter__(self)
    # __next__(self)

    # Overrides
    # -------------------

    def get_children(self, lvl=0, explain=False, leaves=False):
        "Return NodeDict childs"
        result = {}
        for name, child in self._nodes.items():
            if isinstance(child, NodeVal):
                if lvl != 0:
                    value = child.get_children(
                        lvl=lvl - 1, explain=explain, leaves=leaves
                    )
                else:
                    value = child

                if leaves:
                    value = value or child

                result[name] = value

        if explain:
            return {
                "obj": self,
                f"subtype (auto:{self._node_autoconf})": self.conf_children,
                "value": self.get_value(),
                "children": result,
            }
        return result

    def get_value(self, lvl=0, explain=False):
        "Return NodeDict value"

        payload = dict(self._node_conf_parsed)
        children = self._nodes

        for item_def in self._node_conf_struct:

            key = item_def.key
            attr = item_def.attr
            child = children.get(attr)

            if isinstance(child, NodeVal):
                if lvl == 0:
                    if key in payload:
                        del payload[key]
                    if attr in payload:
                        del payload[attr]
                else:
                    payload[attr] = child.get_value(lvl=lvl - 1)

        return payload

    def _node_conf_defaults(self, payload):
        """Return payload merged with default value (NodeDict)"""

        payload = payload or {}
        conf_default = self.conf_default or {}

        # Payload sanity check
        if not isinstance(payload, dict):
            raise DictExpected(f"A dict was expected for {self}, got: {payload}")
        if not isinstance(conf_default, dict):
            raise DictExpected(
                f"A dict was expected for {self}/conf_default, got: {conf_default}"
            )

        # Update payload
        result = copy.deepcopy(conf_default)
        result.update(payload)
        payload = result

        # Init item constructor from current payload
        self._node_conf_struct = NodeDictItemManager(
            self.conf_children,
            payload=payload,
            autoconf=self._node_autoconf,
            log=self._node_log,
        )
        payload = self._node_conf_struct.clean(payload)

        return payload

    def _node_conf_build(self):
        "For NodeDict"

        payload = self._node_conf_parsed

        # Developper sanity check
        if not isinstance(payload, dict):
            # Developper must call node_hook_conf() first if payload are not dict
            raise DictExpected(
                f"A dictionnary was expected for {self}, got {type(payload).__name__}: {payload}"
            )

        self._nodes = {}
        self._node_conf_struct.build_it(self)

    def add_child(self, ident, obj):
        "Add a child node"

        # We always check only NOdeVal object can be added as child
        assert isinstance(
            obj, NodeVal
        ), f"Cannot add non child object to {self}: got {obj}"

        self._nodes[ident] = obj


# NodeMap
# =====================================


class NodeMap(NodeDict):
    "A nodeDict accessible via its attributes"

    def add_child(self, ident, obj):
        "Add a child node"

        # pylint: disable=super-with-arguments
        super(NodeMap, self).add_child(ident, obj)
        setattr(self, ident, obj)
        #print ("SET ATTR", self, ident, obj)

    def __getattr__(self, key):
        """Fetch attribute from config nodes"""

        # TODO:
        # Linter reports an issue when accessing attributes
        # not defined at the class/instance level. How to to fix this?

        if key in self._nodes:
            # print (f"Get value: {key} for {id(self)} from _nodes")
            return self._nodes[key]

        if key in self._node_conf_parsed:
            # print (f"Get value: {key} for {id(self)} from _conf_parsed")
            return self._node_conf_parsed[key]

        # TODO: Implement nice warning for dropped childrens !!!
        # if self.conf_children:
        #     matches = [ remap.attr for remap in self.conf_children if remap.attr]
        #     self.conf_children
        #     print ("NodeDict Struct2")
        #     return self._node_conf_build_future(payload)

        raise AttributeError(f"Missing attribute: {key} in {self}")

    def __setattr__(self, key, value):

        if key in self._nodes:
            # Set attribute if in _nodes
            # print (f"Set node value: {key}={value} for {self}")
            self._nodes[key] = value
            # self.__dict__[key] = value
        elif key in self._node_conf_parsed:
            # print (f"Set conf value: {key}={value} for {self}")
            self._node_conf_parsed[key] = value
        else:
            # or just set regular attribute
            # print (f"Set attr value: {key}={value} for {self}")
            # pylint: disable=super-with-arguments
            super(NodeMap, self).__setattr__(key, value)


# NodeMapEnv
# =====================================


def expand_envar_syntax(payload, cls):
    "Expand serialized dict/list from env syntax"

    if not cls:
        return payload
    if isinstance(payload, cls):
        return payload
    if not isinstance(payload, str):
        return payload

    result = payload
    if cls == dict:
        result = {}
        for statement in payload.split(","):

            keyval = statement.split("=", 1)
            if len(keyval) != 2:
                raise InvalidSyntax(
                    f"Invalid syntax, expected 'key=value', got: {statement}"
                )
            key = keyval[0]
            value = keyval[1]
            result[key] = value

    if cls == list:
        result = []
        for statement in payload.split(","):
            result.append(statement)

    return result


class NodeMapEnv(NodeMap):
    "Like a NodeMap, but fetch value from env"

    conf_env_prefix = None

    def _node_conf_defaults(self, payload):
        """Override payload from environment vars"""

        # pylint: disable=super-with-arguments
        result = super(NodeMapEnv, self)._node_conf_defaults(payload)
        conf_struct = self._node_conf_struct

        # Override from environment
        for key, val in result.items():
            value = self.get_env_conf(key) or val

            cls = [item.cls for item in conf_struct if item.key == key]
            if len(cls) > 0:
                value = expand_envar_syntax(value, cls[0])

            result[key] = value

        return result

    def get_env_conf(self, key=None):
        "Return the value of environment var associated to this object"
        # name = f"{self.module}_{self.kind}_{self.ident}"
        # name = f"{self.kind}_{self.ident}"
        name = f"{self.conf_env_prefix or self.kind}"

        if key:
            name += f"_{key}"
        name = name.upper()
        result = os.getenv(name)

        if result:
            self._node_log.info(f"    3.2 Fetch value from env: {name}={result}")
        else:
            self._node_log.debug(f"    3.2 Skip value from env: {name}")

        return result


# Test decorators
# =====================================


def makemap(cls):
    "Test decorator to create NodeMap classes"

    class Class(cls, NodeMap):
        "Generated class"

    return Class


# Test Class Data
# =====================================


class NodeAuto:
    """
    Autoconfiguration Configuration

    This class will autodetermine what kind of object and apply magically
    the associated classe upon is source object type.
    """

    def __init__(self, *args, ident=None, payload=None, autoconf=-1, **kwargs):

        # Map json object to Node class
        self.__class__ = map_node_class(payload)

        # Forward to class
        # pylint: disable=non-parent-init-called
        self.__init__(*args, ident=ident, payload=payload, autoconf=autoconf, **kwargs)



"""
Cafram utils
"""

# pylint: disable=line-too-long
# pylint: disable=invalid-name

import io
import os

import logging
import json
import re
from io import StringIO

# from pprint import pprint
# from pathlib import Path

import ruamel.yaml
import jsonschema
from jsonschema import Draft202012Validator, validators
import sh


# =====================================================================
# Init
# =====================================================================

# Usage of get_logger:
# # In main app:
#   from paasify.common import get_logger
#   log, log_level = get_logger(logger_name="paasify")
# # In other libs:
#   import logging
#   log = logging.getLogger(__name__)

log = logging.getLogger(__name__)


# Setup YAML object
yaml = ruamel.yaml.YAML()
yaml.version = (1, 1)
yaml.default_flow_style = False
# yaml.indent(mapping=3, sequence=2, offset=0)
yaml.allow_duplicate_keys = True
yaml.explicit_start = True


# =====================================================================
# Logging helpers
# =====================================================================


# Source: https://stackoverflow.com/questions/2183233/how-to-add-a-custom-loglevel-to-pythons-logging-facility/35804945#35804945
def addLoggingLevel(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        raise AttributeError(f"{levelName} already defined in logging module")
    if hasattr(logging, methodName):
        raise AttributeError(f"{methodName} already defined in logging module")
    if hasattr(logging.getLoggerClass(), methodName):
        raise AttributeError(f"{methodName} already defined in logger class")

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):

        if self.isEnabledFor(levelNum):
            # Monkey patch for level below 10, dunno why this not work
            lvl = levelNum
            if levelNum < 10:  # Why this ? :`(
                lvl = 10
                message = f"({levelName}) {message}"
            # pylint: disable=protected-access
            self._log(lvl, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)


class MultiLineFormatter(logging.Formatter):
    """Multi-line formatter."""

    def get_header_length(self, record):
        """Get the header length of a given record."""
        return len(
            super().format(
                logging.LogRecord(
                    name=record.name,
                    level=record.levelno,
                    pathname=record.pathname,
                    lineno=record.lineno,
                    msg="",
                    args=(),
                    exc_info=None,
                )
            )
        )

    def format(self, record):
        """Format a record with added indentation."""
        indent = " " * self.get_header_length(record)
        head, *trailing = super().format(record).splitlines(True)
        return head + "".join(indent + line for line in trailing)


def get_logger(logger_name=None, create_file=False, verbose=None, sformat='default', tformat='default'):
    """Create CmdApp logger"""

    # Take default app name
    logger_name = logger_name or __name__

    # Manage logging level
    if not verbose:
        loglevel = logging.getLogger().getEffectiveLevel()
    else:
        try:
            # pylint: disable=protected-access
            loglevels = list(logging._nameToLevel)
            loglevel = loglevels[verbose]
        except IndexError:
            loglevel = len(loglevels) - 1

    # Create logger for prd_ci
    _log = logging.getLogger(logger_name)
    _log.setLevel(level=loglevel)

    # Formatters
    # See: https://docs.python.org/3/library/logging.html#logrecord-attributes
    sformats = {
        "default": "%(levelname)8s: %(message)s",
        "struct": "%(name)-40s%(levelname)8s: %(message)s",
        "time": "%(asctime)s.%(msecs)03d|%(name)-16s%(levelname)8s: %(message)s",
        "precise": (
            "%(asctime)s.%(msecs)03d"
            + " (%(process)d/%(thread)d) "
            + "%(pathname)s:%(lineno)d:%(funcName)s"
            + ": "
            + "%(levelname)s: %(message)s"
        ),
        }
    tformats = {
        "default": "%H:%M:%S",
        "precise": "%Y-%m-%d %H:%M:%S",
        }

    # formatter = logging.Formatter(format4, tformat1)
    formatter = MultiLineFormatter(sformats[sformat], tformats[tformat])

    # Create console handler for logger.
    stream = logging.StreamHandler()
    # stream.setLevel(level=logging.DEBUG)
    stream.setFormatter(formatter)
    _log.addHandler(stream)

    # Create file handler for logger.
    if isinstance(create_file, str):
        handler = logging.FileHandler(create_file)
        handler.setLevel(level=logging.DEBUG)
        handler.setFormatter(formatter)
        _log.addHandler(handler)

    # print (f"Fetch logger name: {logger_name} (level={loglevel})")

    # Return objects
    return _log


# =====================================================================
# Misc functions
# =====================================================================

# pylint: disable=redefined-builtin
def truncate(data, max=72, txt=" ..."):
    "Truncate a text to max lenght and replace by txt"
    data = str(data)
    if max < 0:
        return data
    if len(data) > max:
        return data[: max + len(txt)] + txt
    return data


def merge_dicts(dict1, dict2):
    """Given two dictionaries, merge them into a new dict as a shallow copy.

    Compatibility for Python 3.5 and above"""
    # Source: https://stackoverflow.com/a/26853961/2352890
    result = dict1.copy()
    result.update(dict2)
    return result


# TODO: Add tests on this one
def to_domain(string, sep=".", alt="-"):
    "Transform any string to valid domain name"

    domain = string.split(sep)
    result = []
    for part in domain:
        part = re.sub("[^a-zA-Z0-9]", alt, part)
        part.strip(alt)
        result.append(part)

    return ".".join(result)


# TODO: Add tests on this one
def first(array):
    "Return the first element of a list or None"
    # return next(iter(array))
    array = list(array) or []
    result = None
    if len(array) > 0:
        result = array[0]
    return result


# TODO: add tests
def from_yaml(string):
    "Transform YAML string to python dict"
    return yaml.load(string)


# TODO: add tests
def to_yaml(obj, headers=False):
    "Transform obj to YAML"
    options = {}
    string_stream = StringIO()

    if isinstance(obj, str):
        obj = json.loads(obj)

    yaml.dump(obj, string_stream, **options)
    output_str = string_stream.getvalue()
    string_stream.close()
    if not headers:
        output_str = output_str.split("\n", 2)[2]
    return output_str


# TODO: add tests
def to_json(obj, nice=True):
    "Transform JSON string to python dict"
    if nice:
        return json.dumps(obj, indent=2)
    return json.dumps(obj)


# TODO: add tests
def from_json(string):
    "Transform JSON string to python dict"
    return json.loads(string)


# TODO: add tests
def to_dict(obj):
    """Transform JSON obj/string to python dict

    Useful to transofmr nested dicts as well"""
    if not isinstance(obj, str):
        obj = json.dumps(obj)
    return json.loads(obj)


def serialize(obj, fmt="json"):
    "Serialize anything, output json like compatible (destructive)"

    # pylint: disable=unnecessary-lambda
    obj = json.dumps(obj, default=lambda o: str(o), indent=2)

    if fmt in ["yaml", "yml"]:
        # Serialize object in json first
        obj = json.loads(obj)

        # Convert json to yaml
        string_stream = io.StringIO()
        yaml.dump(obj, string_stream)
        output_str = string_stream.getvalue()
        string_stream.close()

        # Remove 2 first lines of output
        obj = output_str.split("\n", 2)[2]

    return obj


def duplicates(_list):
    """Check if given list contains duplicates"""
    known = set()
    dup = set()
    for item in _list:
        if item in known:
            dup.add(item)
        else:
            known.add(item)

    if len(dup) > 0:
        return list(dup)
    return []


def read_file(file):
    "Read file content"
    with open(file, encoding="utf-8") as _file:
        return "".join(_file.readlines())


def write_file(file, content):
    "Write content to file"

    file_folder = os.path.dirname(file)
    if not os.path.exists(file_folder):
        os.makedirs(file_folder)

    with open(file, "w", encoding="utf-8") as _file:
        _file.write(content)


def flatten(array):
    "Flatten any arrays nested arrays"
    if array == []:
        return array
    if isinstance(array[0], list):
        return flatten(array[0]) + flatten(array[1:])
    return array[:1] + flatten(array[1:])


# =====================================================================
# JSON Schema framework
# =====================================================================


def _extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for prop, subschema in properties.items():
            if "default" in subschema:
                instance.setdefault(prop, subschema["default"])

        for error in validate_properties(
            validator,
            properties,
            instance,
            schema,
        ):
            yield error

    return validators.extend(
        validator_class,
        {"properties": set_defaults},
    )


DefaultValidatingValidator = _extend_with_default(Draft202012Validator)


def json_validate_defaults(schema, payload):
    "Validate dict against schema and set defaults"
    DefaultValidatingValidator(schema).validate(payload)
    return payload


def json_validate(schema, payload):
    "Validate dict against schema"
    jsonschema.validate(payload, schema)
    return payload


# =====================================================================
# Command Execution framework
# =====================================================================


def _exec(command, cli_args=None, logger=None, **kwargs):
    "Execute any command"

    # Check arguments
    cli_args = cli_args or []
    assert isinstance(cli_args, list), f"_exec require a list, not: {type(cli_args)}"

    # Prepare context
    sh_opts = {
        # "_in": sys.stdin,
        # "_out": sys.stdout,
    }
    sh_opts = kwargs or sh_opts

    # Bake command
    cmd = sh.Command(command)
    cmd = cmd.bake(*cli_args)

    # Log command
    if logger:
        cmd_line = [f"{key}='{val}'" for key, val in sh_opts.get("_env", {}).items()]
        # pylint: disable=protected-access
        cmd_line = (
            cmd_line
            + [cmd.__name__]
            + [x.decode("utf-8") for x in cmd._partial_baked_args]
        )
        cmd_line = " ".join(cmd_line)
        logger.exec(cmd_line)  # Support exec level !!!

    # Execute command via sh
    try:
        output = cmd(**sh_opts)
        return output

    except sh.ErrorReturnCode as err:
        # log.error(f"Error while running command: {command} {' '.join(cli_args)}")
        # log.critical (f"Command failed with message:\n{err.stderr.decode('utf-8')}")

        # pprint (err.__dict__)
        # raise error.ShellCommandFailed(err)
        # sys.exit(1)
        raise err



# -*- coding: utf-8 -*-
"""
Stack components class

"""

import os
from pprint import pprint, pformat  # noqa: F401

import json
import _jsonnet
import anyconfig

from cafram.nodes import NodeList, NodeMap
from cafram.utils import (
    to_json,
    to_yaml,
    write_file,
    first,
    _exec,
)

from paasify.framework import PaasifyObj
import paasify.errors as error


# =======================================================================================
# Stack Assembler
# =======================================================================================


class StackDumper(PaasifyObj):
    "StackDumper for dumping data into files for troubleshooting purpose"

    def __init__(self, path, enabled=True):

        self.out_dir = path
        self.enabled = enabled

        if self.enabled:
            self.cleanup()

    def cleanup(self):
        "Cleanup target directory"
        path = self.out_dir

        if not os.path.exists(path):
            self.log.info(f"StackDumper created directory: {path}")
            os.makedirs(path)
        for file_ in os.listdir(path):
            rm_file = os.path.join(path, file_)
            self.log.notice(f"StackDumper removed file: {rm_file}")
            os.remove(rm_file)

    def dump(self, file_name, content, fmt=None):
        "Dump any data"

        if not self.enabled:
            return

        dest = os.path.join(self.out_dir, file_name)

        if fmt == "json":
            content = to_json(content)
        elif fmt in ["yml", "yaml"]:
            content = to_yaml(content)
        elif fmt in ["pprint", "pformat"]:
            content = pformat(content) + "\n"

        if not isinstance(content, str):
            content = str(content)

        self.log.info(f"Dumping data into: {dest}")
        write_file(dest, content)

    def show_diff(self):
        "Show a colored diff outpout between files"

        if not self.enabled:
            return

        print("==== Dump differential jsonnet output")
        path = os.path.realpath(self.out_dir)
        prev = os.path.join(path, "1-docker-compose.yml")
        print(_exec("tail", cli_args=["-n", "9999", prev]))
        for file_ in sorted(os.listdir(path)):
            if file_.startswith("2-") and file_.endswith("out.yml"):
                file_ = os.path.join(path, file_)
                opts = [
                    "--color=always",
                    "-u",
                    prev,
                    file_,
                ]
                out = _exec("diff", cli_args=opts, _ok_code=[0, 1])
                print(out)
                prev = file_


# =======================================================================================
# Stack Assembler
# =======================================================================================


class StackAssembler(PaasifyObj):
    "Object to manage stack assemblage"

    # Internal object:
    # all_tags
    # engine

    # Docker processors
    # ===========================
    def _get_docker_files(self, all_tags):
        "Retrieve the list of tags docker-files"

        # TODO: Deprecated this, this has already been done before somewhere in the process !

        docker_files = []
        for cand in all_tags:
            docker_file = cand.get("docker_file")
            if docker_file:
                docker_files.append(docker_file)
                self.log.info(f"Insert docker-compose: {docker_file}")

        return docker_files

    def assemble_docker_compose(
        self, all_tags, engine, env=None, dump_payload_log=False
    ):
        "Generate the docker-compose file"

        docker_files = self._get_docker_files(all_tags)

        # Report to user
        env = env or {}
        assert isinstance(env, dict), f"Got: {env}"

        if dump_payload_log:
            self.log.trace("Docker vars:")
            for key, val in sorted(env.items()):
                self.log.trace(f"  {key}: {val}")

        out = engine.assemble(docker_files, env=env)
        # Exception is too wide !
        # try:
        #    out = engine.assemble(docker_files, env=env)
        # except Exception as err:
        # # TOTEST => except sh.ErrorReturnCode as err:
        #    err = bin2utf8(err)
        #    # pylint: disable=no-member
        #    self.log.critical(err.txterr)
        #    raise error.DockerBuildConfig(
        #        f"Impossible to build docker-compose files: {err}"
        #    ) from err

        # Fetch output
        docker_run_content = out.stdout.decode("utf-8")
        docker_run_payload = anyconfig.loads(docker_run_content, ac_parser="yaml")
        return docker_run_payload

    # Vars processors
    # ===========================

    def process_jsonnet_exec(self, file, action, data, import_dirs=None):
        "Process jsonnet file"

        # Developper init
        import_dirs = import_dirs or []
        data = data or {}
        assert isinstance(data, dict), f"Data must be dict, got: {data}"
        # assert len(import_dirs) > 2, f"Missing import dirs, got: {import_dirs}"

        # TODO: Enforce jsonnet API
        # assert action in [
        #     "metadata",
        #     "vars_default",
        #     "vars_override",
        #     "process_globals", # Testing WIP
        #     "process_transform", # Testing WIPP
        #     "docker_override",
        # ], f"Action not supported: {action}"

        # Prepare input variables
        mod_ident = os.path.splitext(os.path.basename(file))[0]
        ext_vars = {
            "parent": json.dumps(mod_ident),
            "action": json.dumps(action),
        }
        for key, val in data.items():
            ext_vars[key] = json.dumps(val)

        def try_path(dir_, rel):
            "Helper function to load a jsonnet file into memory for _jsonnet"

            if not rel:
                return None, None
                # raise RuntimeError("Got invalid filename (empty string).")

            if rel[0] == "/":
                full_path = rel
            else:
                full_path = os.path.join(dir_, rel)

            if full_path[-1] == "/":
                return None, None
                # raise RuntimeError("Attempted to import a directory")

            if not os.path.isfile(full_path):
                return full_path, None
            with open(
                full_path,
                encoding="utf-8",
            ) as file_:
                return full_path, file_.read()

        # Jsonnet import callback
        def import_callback(dir_, rel):
            "Helper function to load a jsonnet libraries in lookup paths"

            test_dirs = [dir_] + import_dirs
            for test_dir in test_dirs:
                full_path, content = try_path(test_dir, rel)
                self.log.trace(f"Load '{rel}' jsonnet from: {full_path}")
                if content:
                    return full_path, content.encode()

            test_dirs = " ".join(test_dirs)
            raise RuntimeError(
                f"Jsonnet file not found '{rel}' in any of these paths: {test_dirs}"
            )

        # Process jsonnet tag
        self.log.trace(f"Process jsonnet: {file} (action={action})")
        try:
            # pylint: disable=c-extension-no-member
            result = _jsonnet.evaluate_file(
                file,
                ext_vars=ext_vars,
                import_callback=import_callback,
            )
        except RuntimeError as err:
            self.log.critical(f"Can't parse jsonnet file: {file}")
            raise error.JsonnetBuildFailed(err)

        # Return python object from json output
        result = json.loads(result)
        return result


# =======================================================================================
# Stack Apps
# =======================================================================================


class StackApp(NodeMap, PaasifyObj):
    "Stack Applicationk Object"

    # conf_logger = "paasify.cli.stack_app"

    conf_default = {
        "app": None,
        "app_source": None,
        "app_path": None,
        "app_name": None,
    }

    @property
    def name(self):
        "App name attribute"
        return self.app_name

    # def node_hook_init(self, **kwargs):
    def node_hook_init(self):

        # Future: let's propagate from the top instead ...
        # parents = self.get_parents()
        # stack = parents[1]
        # prj = parents[3]
        self.stack = self._node_parent
        self.prj = self.stack._node_parent._node_parent
        self.sources = self.prj.sources

        self.app_dir = None

    def node_hook_transform(self, payload):

        if isinstance(payload, str):
            payload = {"app": payload}

        app_def = payload.get("app")
        app_path = payload.get("app_path")
        app_source = payload.get("app_source")
        app_name = payload.get("app_name")

        app_split = app_def.split(":", 2)

        if len(app_split) == 2:
            app_source = app_source or app_split[0] or "default"
            app_path = app_path or app_split[1]
        else:
            # Get from default namespace
            app_name = app_source or app_split[0] or "default"
            app_source = "default"
            app_path = app_name
        app_def = f"{app_source}:{app_path}"

        if not app_name:
            app_name = os.path.split(app_path)[-1]

        result = {
            "app": app_def,
            "app_path": app_path,
            "app_source": app_source,
            "app_name": app_name,
        }

        return result

    def get_app_source(self):
        "Return app source"

        app_source = self.app_source or None
        src = self.prj.sources.get_app_source(self.name, source=app_source)
        return src

    def get_app_path(self):
        "Return app path"

        src = self.get_app_source()
        ret = os.path.join(src.path, self.name)
        return ret


# =======================================================================================
# Stack Tag schemas
# =======================================================================================

stack_name_pattern = {
    "title": "Short form",
    "description": (
        "Just pass the tag you want to apply as string."
        " This form does not allow jsonnet ovar override"
    ),
    "type": "string",
    "oneOf": [
        {
            "title": "Reference collection app",
            "description": (
                "Reference a tag from a specific collection."
                " This form does not allow jsonnet ovar override"
            ),
            "pattern": "^.*:.*$",
        },
        {
            "title": "Direct or absolute app path",
            "description": ("Reference a tag from a specific collection."),
            "pattern": ".*/[^:]*",
        },
        {
            "title": "Tag",
            "description": ("Will find the best matvhing tag."),
            "pattern": "^.*$",
        },
    ],
}


stack_ref_kind_defs = [
    {
        "title": "With value",
        "description": "Pass extra vars for during jsonet tag processing.",
        "type": "object",
    },
    {
        "title": "Without value",
        "description": ("No vars are added for this jsonnet tag processing."),
        "type": "null",
    },
]
stack_ref_defs = {
    "[~].*": {
        "title": "Disabled Tag: ~$tag_name",
        "description": (
            "Completely disable a tag from the processing list, whatever how many instances there are."
        ),
        "oneOf": stack_ref_kind_defs,
        "default": {},
    },
    "^.*:.*$": {
        "title": "Collection tag: $collection_name:$tag_name",
        "description": (
            "Reference a tag from a specific collection."
            "See: Specific tag documentation for further informations."
        ),
        "oneOf": stack_ref_kind_defs,
        "default": {},
    },
    # ".*/[^:]*": {
    #     "title": "Absolute tag: $tag_path",
    #     "description": (
    #         "Reference a tag from a absolute app path."
    #     ),
    # },
    ".*": {
        "title": "Tag name: $tag_name",
        "description": (
            "Will find the best matching tag."
            "See: Specific tag documentation for further informations."
        ),
        "oneOf": stack_ref_kind_defs,
        "default": {},
    },
}


class StackTag(NodeMap, PaasifyObj):
    """Paasify Stack object"""

    conf_schema = {
        # "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "StackTag configuration",
        "description": (
            "Tag definition. It support two formats at the same time: as string or dict."
            " If the name is prefixed with a `!`, then it is removed from the"
            " processing list (both vars, docker-file and jsonnet processing)."
        ),
        "oneOf": [
            {
                "title": "As string",
                "description": (
                    "Just pass the tag you want to apply as string."
                    " This form does not allow jsonnet ovar override"
                ),
                "type": "string",
                "default": "",
                "examples": [
                    {
                        "tags": [
                            "my_tagg",
                            "~my_prefix_tag",
                            "my_collection:my_prefix_tag",
                        ],
                    },
                ],
                "oneOf": [
                    {
                        "title": stack["title"],
                        "description": stack["description"],
                        "pattern": stack_pattern,
                    }
                    for stack_pattern, stack in stack_ref_defs.items()
                ],
            },
            {
                "title": "As object",
                "description": (
                    "Define a tag. The key represent the name of the"
                    " tag, while it's value is passed as vars during"
                    " jsonnet processing. This form allow jsonnet ovar override"
                ),
                "type": "object",
                "default": {},
                "examples": [
                    {
                        "tags": [
                            {
                                "other_tag": {
                                    "specific_conf": "val1",
                                }
                            },
                            {"my_collection:another_tag": None},
                            {
                                "~ignore_this_tag": {
                                    "specific_conf": "val1",
                                }
                            },
                        ],
                    },
                ],
                "minProperties": 1,
                "maxProperties": 1,
                # "additionalProperties": False,
                "patternProperties": stack_ref_defs,
            },
        ],
    }

    conf_ident = "{self.name}={self.vars}"

    conf_children = [
        {
            "key": "name",
            "cls": str,
        },
        {
            "key": "vars",
        },
    ]

    # Place to store list of candidates
    jsonnet_candidates = None
    docker_candidates = None

    # Object shortcuts
    stack = None
    prj = None
    app = None

    def node_hook_transform(self, payload):

        # Init parent objects
        self.stack = self.get_parent().get_parent()
        self.prj = self.stack.get_parent().get_parent()
        self.app = self.prj.get_parents()

        # Transform input
        result = {
            "name": None,
            "vars": {},
        }
        if isinstance(payload, str):
            result["name"] = payload

        elif isinstance(payload, dict):

            keys = list(payload.keys())
            if len(keys) == 1:
                # TODO: replace this by a common function !
                for key, val in payload.items():
                    result["name"] = key
                    result["vars"] = val
            elif len(keys) == 0:
                raise Exception(f"Missing tag name: {payload}")
            else:
                result.update(payload)

        else:
            raise Exception(f"Not supported type: {payload}")

        assert result["name"]
        return result


class StackTagMgr(NodeList, PaasifyObj):
    "Manage Stack Tags"

    conf_schema = {
        # "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "Paasify Stack Tags configuration",
        "description": "Determine a list of tags to apply",
        "oneOf": [
            {
                "title": "List of tags",
                "description": (
                    "Define a list of tags. You can interact in few ways with"
                    " tags. Tags can support boths syntaxes at the same time."
                ),
                "type": "array",
                "default": [],
                "additionalProperties": StackTag.conf_schema,
                # "items": StackTag.conf_schema,
                "examples": [
                    {
                        "tags": [
                            "my_tagg",
                            "~my_prefix_tag",
                            "my_collection:my_prefix_tag",
                            {
                                "other_tag": {
                                    "specific_conf": "val1",
                                }
                            },
                            {"my_collection:another_tag": None},
                            {
                                "~ignore_this_tag": {
                                    "specific_conf": "val1",
                                }
                            },
                        ],
                    },
                ],
            },
            {
                "title": "Unset",
                "description": "Do not declare any tags",
                "type": "null",
                "default": None,
                "examples": [
                    {
                        "tags": None,
                    },
                ],
            },
        ],
    }

    conf_children = StackTag
    module = "paasify.cli"

    def node_hook_transform(self, payload):

        # Fetch tag lists
        tag_stack_list = payload["raw"]
        tag_prefix = payload["tag_prefix"]
        tag_suffix = payload["tag_suffix"]

        # Process keys (before parsing :/ )
        tag_stack_keys = [tag for tag in tag_stack_list if isinstance(tag, str)]
        tag_stack_keys = tag_stack_keys + [
            first(tag.keys()) for tag in tag_stack_list if isinstance(tag, dict)
        ]

        # Remove duplicates tags
        tag_items = []
        for item in tag_prefix:
            if item not in tag_stack_keys:
                tag_items.append(item)
        tag_items.extend(tag_stack_list)
        for item in tag_suffix:
            if item not in tag_stack_keys:
                tag_items.append(item)

        return tag_items

    def node_hook_final(self):
        "Remove disabled tags"

        markers = "^"
        excluded_tags = [
            tag.name for tag in self._nodes if tag.name.startswith(markers)
        ]
        if excluded_tags:
            self.log.debug(f"Disabling tags: {excluded_tags}")
        for excluded in excluded_tags:
            clean = excluded[1:]
            xclude = [excluded, clean]
            self._nodes = [tag for tag in self._nodes if tag.name not in xclude]



# -*- coding: utf-8 -*-
"""
Core Paasify Library
"""

import logging
from pathlib import Path
from single_version import get_version
from cafram.utils import addLoggingLevel

# Add logging levels for the whole apps
addLoggingLevel("NOTICE", logging.INFO + 5)
addLoggingLevel("EXEC", logging.DEBUG + 5)
addLoggingLevel("TRACE", logging.DEBUG - 5)

# Manage version from poetry
__version__ = get_version("paasify", Path(__file__).parent.parent)



# -*- coding: utf-8 -*-
"""
Paasify Stack management

This library provides two classes:

* StackManager: Manage a list of stacks
* Stack: A stack instance
"""

# pylint: disable=logging-fstring-interpolation


import os

import re
from pprint import pprint, pformat  # noqa: F401
from functools import wraps

import anyconfig


from cafram.nodes import NodeList, NodeMap
from cafram.utils import (
    to_domain,
    to_yaml,
    first,
    write_file,
    to_json,
)

import paasify.errors as error
from paasify.common import uniq
from paasify.vars import VarMgr
from paasify.framework import (
    PaasifyObj,
    PaasifyConfigVars,
    FileLookup,
    FileReference,
)
from paasify.stack_components import (
    StackTagMgr,
    StackApp,
    StackAssembler,
    StackDumper,
)


# Try to load json schema if present
ENABLE_JSON_SCHEMA = False
try:
    from json_schema_for_humans.generate import generate_from_filename
    from json_schema_for_humans.generation_configuration import GenerationConfiguration

    ENABLE_JSON_SCHEMA = True
except ImportError:
    ENABLE_JSON_SCHEMA = False


class Stack(NodeMap, PaasifyObj):
    "Paasify Stack Instance"

    conf_default = {
        "dir": None,
        "name": None,
        "app": None,
        "tags": [],
        "tags_suffix": [],
        "tags_prefix": [],
        "vars": [],
    }

    conf_children = [
        {
            "key": "app",
            "cls": StackApp,
            "action": "unset",
            "hook": "node_hook_app_load",
        },
        {
            "key": "vars",
            "cls": PaasifyConfigVars,
        },
    ]

    conf_schema = {
        # TODO: Bug: We want a way in stacks to have random/useless values, because
        # while doing trial/errors, the parser becomes annoying. Ie, if extra key
        # starts with a `_`, then simply skip the parser error. This could be a switch
        # like `--develop`
        "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "Paasify Stack configuration",
        "default": conf_default,
        "oneOf": [
            {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "name": {
                        "type": "string",
                        "minLength": 2,
                        "title": "Name of the stack",
                        "description": "This is a stack ident, it should not be changed. If not set, it comes from the app name, and from the dir if no apps. Should not contains any special char, except - (or _). This is used to name the app (docker, dns, etc ...)",
                    },
                    "dir": {
                        "type": "string",
                        "title": "Directory where live the stack",
                        "description": "Directory where the docker-compose.run.yml file is generated. Can be used to move stack directories (cold move). Default comes from the name.",
                        "default": "$name",
                    },
                    "app": {
                        "type": "string",
                        "title": "Application call",
                        "description": "Shortcut to use an application without modifying it's default parameters. First optional part is module collection, and after the ':', it's the name/path to the aplication inside the module",
                    },
                    "tags": StackTagMgr.conf_schema,
                    "tags_prefix": StackTagMgr.conf_schema,
                    "tags_suffix": StackTagMgr.conf_schema,
                    "vars": PaasifyConfigVars.conf_schema,
                },
            },
            {
                "type": "string",
                "title": "Direct application call",
                "description": "Shortcut to use an application without modifying it's default parameters. First optional part is module collection, and after the ':', it's the name/path to the aplication inside the module",
                "pattern": "^([^:]+:)?(.*)$",
            },
        ],
    }

    # Children objects
    tag_manager = None
    engine = None
    prj = None

    # Stack vars
    stack_dir = None
    stack_name = None
    prj_dir = None
    prj_ns = None

    # Vars scopes
    var_scopes = {
        "global": ["default_conf", "extra_vars_yaml", "global_conf"],
        "stack": [
            "default_conf",
            "app_yaml",
            "stack_yaml",
            "extra_vars_yaml",
            "global_conf",
            "tag_def",
            "tag_dyn",
            "stack_conf",
        ],
    }

    # CaFram functions
    # ---------------------

    def node_hook_init(self):
        "Create instance attributes"

        self._cache = {}

        # Internal attributes
        self.prj = self.get_parent().get_parent()
        self.runtime = self.prj.runtime
        assert (
            self.prj.__class__.__name__ == "PaasifyProject"
        ), f"Expected PaasifyProject, got: {self.prj}"

    def node_hook_transform(self, payload):

        if isinstance(payload, str):
            key = "app" if ":" in payload else "name"
            payload = {
                key: payload,
            }

        assert isinstance(payload, dict)
        return payload

    def node_hook_app_load(self):
        "Modify stack depending app"

        # Assert name,dir,app
        stack_name = self.name
        stack_dir = self.dir
        stack_app = self.app

        # Check name first
        if not stack_name:

            # Fetch from app
            if stack_app:
                stack_name = stack_app.app_name

            # Fetch from dir
            elif stack_dir:
                # stack_name = os.path.split(stack_dir)[-1]
                stack_name = "_".join(stack_dir.split(os.path.sep))

            if not stack_name:
                assert False, f"Missing name, or app or path for stack: {self}"

        if not stack_dir:
            stack_dir = stack_name

        if "/" in stack_name:
            # TODO: Workaround, this should be handled before ...
            stack_name = stack_name.replace("/", "-")

        # Register required vars
        self.stack_name = stack_name
        self.stack_dir = stack_dir
        self.prj_ns = self.prj.config.namespace or self.prj.runtime.namespace
        self.prj_path = self.prj.runtime.root_path
        self.stack_path = os.path.join(self.prj.runtime.root_path, stack_dir)
        self.stack_dump_path = os.path.join(
            self.runtime.project_private_dir, "dumps", stack_dir
        )
        self.ident = self.stack_name
        self.stack_path_abs = os.path.abspath(self.stack_path)

        # Check
        assert self.stack_name, f"Bug here, should not be empty, got: {self.stack_name}"
        assert re.search("^[a-zA-Z0-9_].*$", self.stack_name), f"Got: {self.stack_name}"
        assert re.search("^[a-zA-Z0-9_/].*$", self.stack_dir), f"Got: {self.stack_dir}"

    def node_hook_final(self):
        "Enable CLI debugging"

        # Create engine instance
        payload = {
            "stack_name": f"{self.prj_ns}_{self.stack_name}",
            "stack_path": self.stack_path,
            # os.path.join(self.stack_dir, "docker-compose.run.yml"),
            "docker_file": "docker-compose.run.yml",
        }
        self.engine = self.prj.engine_cls(parent=self, payload=payload)

        # Prepare stack lists
        tag_config = {
            "raw": self.tags or self.prj.config.tags,
            "tag_prefix": ["_paasify"]
            + (self.tags_prefix or self.prj.config.tags_prefix),
            "tag_suffix": self.tags_suffix or self.prj.config.tags_suffix,
        }

        # Start managers
        self.tag_manager = StackTagMgr(
            parent=self, ident=self.stack_name, payload=tag_config
        )
        self.var_manager = None

        self.log.info(f"Stack config: {self.stack_name} in {self.stack_path}")

        # Set sh default args
        self.default_sh_args = {"_fg": True}
        if self.prj.runtime.no_tty:
            self.default_sh_args = {"_in": False}

    # Local functions
    # ---------------------

    def docker_candidates(self) -> list:
        """Return all docker-files candidates: local, app and tags

        Search docker-compose files in the following dirs:

          * Main docker-compose:
            * <local>/docker-compose.y?ml
            * <app>/docker-compose.y?ml
          * Additional docker-composes:
            * <local>/docker-compose.<tag>.y?ml
            * <app>/docker-compose.<tag>.y?ml

        Return the list of candidates for the stack
        """

        # 0. Check cache
        _key_cache = "docker_candidates"
        results = self._cache.get(_key_cache)
        if results:
            return results

        # 1. Init
        app = self.app or None
        pattern = ["docker-compose.yml", "docker-compose.yaml"]

        # 2. Lookup stack docker-compose files
        lookup_app = FileLookup()
        lookup_app.append(self.stack_path, pattern)

        # 3. Lookup app docker-compose files
        if app:
            app_path = app.get_app_path()
            lookup_app.append(app_path, pattern)

        # 4. Assemble results
        matches = lookup_app.match()
        results = [match["match"] for match in matches]

        # 5. Sanity check
        for file in results:
            assert isinstance(file, str), f"Got: {file}"

        # 6. Filter result
        if len(results) < 1:
            paths = [look["path"] for look in lookup_app.get_lookups()]
            paths = ", ".join(paths)
            msg = f"Can't find 'docker-compose.yml' for stack '{self._node_conf_raw}' in: {paths}"
            raise error.StackMissingDockerComposeFile(msg)

        # TODO: Test ideas: test if local_cand and app_cand are properly setup depending the pattern

        # Set in cache and return value
        self._cache[_key_cache] = results
        return results

    def get_tag_plan(self) -> list:
        """
        Resolve all files associated to tags

        Return the list of tags with files
        """

        # 0. Init
        # Objects:
        app = self.app or None
        # Vars:
        stack_dir = self.stack_path
        project_jsonnet_dir = self.prj.runtime.project_jsonnet_dir
        docker_candidates = self.docker_candidates()

        # 1. Generate default tag (docker compose files only)
        tag_base = {
            "tag": None,
            "jsonnet_file": None,
            "docker_file": first(docker_candidates),
        }

        # 2. Forward to StackTagManager: Generate directory lookup for tags

        # Get jsonnet search paths
        dirs_docker = [stack_dir]
        if app:
            src = app.get_app_path()
            assert src
            dirs_docker.append(src)
        dirs_jsonnet = [
            self.prj.runtime.paasify_plugins_dir,
            stack_dir,
            project_jsonnet_dir,
        ]
        for src in self.prj.sources.get_all():
            dirs_jsonnet.append(os.path.join(src.path, ".paasify", "plugins"))

        # Build tag list
        tags = self.tag_manager.get_children()

        tag_list = []
        for tag in tags:

            # Docker lookup
            pattern = [
                f"docker-compose.{tag.name}.yml",
                f"docker-compose.{tag.name}.yaml",
            ]
            lookup = FileLookup()
            for dir_ in dirs_docker:
                lookup.append(dir_, pattern)
            docker_cand = lookup.match()

            docker_file = None
            if len(docker_cand) > 0:
                docker_file = first(docker_cand)["match"]

            # Jsonnet lookup
            pattern = [f"{tag.name}.jsonnet"]
            lookup = FileLookup()
            jsonnet_file = None
            for dir_ in dirs_jsonnet:
                lookup.append(dir_, pattern)
                jsonnet_cand = lookup.match()
                if len(jsonnet_cand) > 0:
                    jsonnet_file = first(jsonnet_cand)["match"]

            ret = {
                "tag": tag,
                "jsonnet_file": jsonnet_file,
                "docker_file": docker_file,
            }
            tag_list.append(ret)

            # Report error to user on missing tags
            if not docker_file and not jsonnet_file:
                msg = f"Can't find '{tag.name}' for '{self.stack_name}' stack"
                raise error.MissingTag(msg)

        # 3. Return result list
        results = []
        results.append(tag_base)
        results.extend(tag_list)
        return results, dirs_jsonnet

    def _gen_conveniant_vars(self, docker_file, tag_names=None) -> dict:
        "Generate default core variables"

        # Extract stack config
        tag_names = tag_names or []
        dfile = anyconfig.load(docker_file, ac_ordered=True, ac_parser="yaml")
        default_service = first(dfile.get("services", ["default"]))
        default_network = first(dfile.get("networks", ["default"]))

        assert isinstance(self.prj_ns, str)
        assert isinstance(self.prj_path, str)
        assert isinstance(self.stack_name, str)
        assert isinstance(self.prj_path, str)
        assert isinstance(self.prj_path, str)
        assert self.stack_path_abs.startswith("/")

        # Build default (only primitives)
        result = {
            "paasify_sep": "-",
            "paasify_sep_dir": os.sep,
            # See: https://www.docker.com/blog/announcing-compose-v2-general-availability/
            "paasify_sep_net": "_",
            "_prj_path": self.prj_path,
            "_prj_namespace": self.prj_ns,  # deprecated because too long !
            "_prj_ns": self.prj_ns,
            "_prj_domain": to_domain(self.prj_ns),
            "_prj_stack_path": self.stack_path,
            # Colon is used here for easier to parsing for later ...
            "_prj_stack_tags": f":{':'.join(tag_names)}:",
            "_stack_name": self.stack_name,
            "_stack_path_abs": self.stack_path_abs,
            "_stack_network": default_network,
            "_stack_service": default_service,
            # To report below as well
            "_stack_app_name": None,
            "_stack_app_dir": None,
            "_stack_app_path": None,
            "_stack_collection_app_path": None,
        }

        app = self.app
        if app:
            extra = {
                "_stack_app_name": os.path.basename(app.app_name),
                "_stack_app_dir": app.app_name,
                "_stack_app_path": app.get_app_path(),
                # TODO: Broken ... use app.src instead ?
                # "_stack_collection_app_path": self.app.collection_dir,
            }
            result.update(extra)
        return result

    @property
    def docker_vars_lookup(self) -> list:
        "Return the lookup configuration for vars.yml location"

        # Lookup config
        fl_stack = {
            "path": self.stack_path,
            "pattern": ["vars.yml", "vars.yaml"],
            "kind": "stack",
            "owner": "user",
        }
        lookup = FileLookup()
        lookup.append(**fl_stack)

        if self.app:
            app_dir = self.app.get_app_path()
            assert app_dir, "Missing app name!"
            fl_app = {
                "path": app_dir,
                "pattern": ["vars.yml", "vars.yaml"],
                "kind": "app",
                "owner": "app",
            }
            lookup.insert(**fl_app)

        return lookup

    @property
    def extra_vars_lookups(self) -> list:
        "Return the lookup configuration for extra_vars location"

        prj_config = self.get_parents()[-2].config
        runtime = self.prj.runtime

        extra_vars = prj_config.extra_vars.get_value()
        extra_vars = extra_vars if isinstance(extra_vars, list) else [extra_vars]

        extra_vars_lookups = FileLookup()
        for ref in extra_vars:
            ref = FileReference(ref, root=runtime.root_path)
            dir_, file_ = os.path.split(ref.path())
            xtra = {
                "kind": "extra_vars",
                "owner": "user",
            }
            extra_vars_lookups.append(dir_, [file_], **xtra)
        return extra_vars_lookups

    def render_vars(
        self,
        hint=None,
        parse=True,
        skip_undefined=False,
        scope=None,
        parse_vars=None,
        varmgr=None,
    ):
        "Return parsed vars for a given scope"

        assert hint, f"Missing valid hint for render_vars call for {self} ..."

        # Detect selector
        parse_vars = parse_vars or {}
        func = None
        scopes = None
        varmgr = varmgr or self.var_manager
        assert varmgr, "Var manager is not initialized or provided !"
        if isinstance(scope, str):
            scopes = scope.split(",")
        elif isinstance(scope, list):
            scopes = scope

        # Attribute function
        def _func(var):
            return var.scope in scopes

        func = _func if scopes else scope

        # Parse the result
        msg = f"Environment rendering asked for scope: parse={parse}, hint={hint}"
        self.log.trace(msg)
        result = varmgr.render_env(
            parse=parse,
            parse_vars=parse_vars,
            skip_undefined=skip_undefined,
            select=func,
            hint=hint,
        )
        return result

    def get_stack_vars(self, sta, all_tags, jsonnet_lookup_dirs, extra_user_vars=None):

        # 0. Get ojects we need to query
        varmgr = VarMgr(parent=self, ident=f"{self.stack_name}")

        globvars = self.prj.config.vars
        localvars = self.vars
        extra_user_vars = extra_user_vars or {}

        # 1. Generate data structures
        docker_file = all_tags[0]["docker_file"]
        tag_names = uniq([tag["tag"].name for tag in all_tags if tag["tag"]])

        vars_default = self._gen_conveniant_vars(
            docker_file=docker_file, tag_names=tag_names
        )
        stack_name = vars_default["_stack_name"]

        vars_global = {var.name: var.value for var in globvars.get_vars_list()}
        vars_user = {var.name: var.value for var in localvars.get_vars_list()}

        # 2. Create a varmgr VarMgr and inject configs

        # Order matters here:
        #   - Add core vars, the one starting with _
        #   - Add extra_vars, from external files
        #   - Add vars from app/vars.yml, then stack/vars.yml
        #   - Add global conf vars from paasify.yml
        #   - Add stack conf vars from paasify.yml
        varmgr.add_vars(
            vars_default,
            scope="global",
            kind="default_conf",
            source="core",
            file="paasify.py",
            owner="paasify",
            parse_order=1000,
        )
        varmgr.add_vars_from_lookup(
            self.extra_vars_lookups, 4000, fail_on_missing=True, scope="global"
        )
        varmgr.add_vars_from_lookup(self.docker_vars_lookup, 5000, scope="stack")
        varmgr.add_vars(
            vars_global,
            scope="global",
            kind="global_conf",
            source="core",
            file="paasify.yml:config.vars",
            owner="user",
            parse_order=7000,
        )
        varmgr.add_vars(
            vars_user,
            scope="stack",
            kind="stack_conf",
            source="core",
            file=f"paasify.yml:stacks[{stack_name}]vars",
            owner="user",
            parse_order=8000,
        )

        # 4. Add config for tags as well
        cand_index = -2  # TODO: we sould remove -2 if app or -1 if not
        tag_instances = []
        prio_index = 0
        for cand in all_tags:
            cand_index += 1
            prio_index += 1

            # Fetch the tag
            # --------------------
            tag = cand.get("tag")
            if not tag:
                continue
            tag_vars = tag.vars or {}
            tag_name = tag.name

            jsonnet_file = cand.get("jsonnet_file")
            if not jsonnet_file:
                # Throw a user warning about wrong config
                if len(tag_vars) > 0:
                    msg = f"Tag vars are only supported for jsonnet tags: {tag_name}: {tag_vars}"
                    self.log.warning(msg)
                continue

            # Prepare loop metadata
            # --------------------
            tag_index = 0
            tag_inst = f"{tag_name}{tag_index}"
            while tag_inst in tag_instances:
                tag_index += 1
                tag_inst = f"{tag_name}{tag_index}"
            tag_instances.append(tag_inst)

            # Assemble context and conf
            # --------------------
            loop_vars = {}
            loop_vars.update(tag_vars)

            varmgr.add_vars(
                loop_vars,
                # scope="tag",
                scope=f"tag_{tag_inst}",
                kind="tag_conf",
                source=tag_inst,
                file=f"paasify.yml:stacks[{stack_name}]tags[{tag_inst}]",
                owner="user",
                parse_order=9000 + prio_index,
            )

            # 3.2 Execute jsonnet plugin var calls
            # --------------------
            if tag_index > 0:
                # We only process ONE time the filter of each kind
                continue

            ctx = self.render_vars(
                scope="global,stack",
                parse=False,
                hint=f"jsonnet scoped vars: {tag_name}",
                varmgr=varmgr,
            )

            tmp2 = sta.process_jsonnet_exec(
                jsonnet_file,
                "plugin_vars",
                {"args": ctx},
                import_dirs=jsonnet_lookup_dirs,
            )
            try:
                var_def = tmp2["def"]
                var_dyn = tmp2["dyn"]
            except KeyError as err:
                self.log.error(
                    f"Could not execute plugin '{jsonnet_file}', please check the plugin is working as expected, especially ensure variable 'metadata.ident' is correctly set to the plugin name"
                )
                self.log.debug(f"Plugin returned: {tmp2}")
                raise error.JsonnetBuildFailed(err)

            varmgr.add_vars(
                var_def,
                scope="stack",
                owner=tag_name,
                kind="tag_def",
                source=tag_name,
                file=jsonnet_file,
                parse_order=2000 + prio_index,
            )
            varmgr.add_vars(
                var_dyn,
                scope="stack",
                owner=tag_name,
                kind="tag_dyn",
                source=tag_name,
                file=jsonnet_file,
                parse_order=3000 + prio_index,
            )

        return varmgr

    def log_extra_payloads(self, payload, msg):
        "Dump to log large chunks of data"

        if not self.dump_data_log:
            return

        self.log.trace("= {msg}")
        self.log.trace("=" * 60)
        self.log.trace("{pformat(payload)}")
        self.log.trace("=" * 60)

    def dump_diff_stage1(self, dumper, varmgr, glob_vars, stack_vars, all_tags):
        "Report build diff stage1"

        txt = []
        txt.append("==== Global Scope: Vars")
        txt.append(
            varmgr.explain(scope=lambda var: var.scope in ["global"], as_string=True)
        )
        txt.append("\n==== Global Scope: Parsed")
        txt.append(to_yaml(glob_vars))
        txt = "\n".join(txt)
        dumper.dump("0-glob-env.txt", txt)

        txt = []
        txt.append("==== Stack Scope: Tags")
        txt.append(pformat(all_tags))
        txt.append("\n==== Stack Scope: Vars")
        txt.append(
            varmgr.explain(
                scope=lambda var: var.scope in ["global", "stack"], as_string=True
            )
        )
        txt.append("\n==== Stack Scope: Parsed")
        txt.append(to_yaml(stack_vars))
        txt = "\n".join(txt)
        dumper.dump("0-stack-env.txt", txt)

    def dump_diff_stage2(
        self, dumper, varmgr, loop_vars, scope_filter, tag_inst, cand_index
    ):
        "Report build diff stage2"
        txt = []
        txt.append("==== Jsonnet Scope: Explain")
        txt.append(varmgr.explain(scope=scope_filter, as_string=True))
        txt.append("\n==== Jsonnet Scope: Vars")
        txt.append(to_yaml(loop_vars))
        txt = "\n".join(txt)

        dumper.dump(f"2-{cand_index:02d}-{tag_inst}-env.txt", txt)

    def assemble(self, vars_only=False, dump_payloads=False, explain=False):
        """Generate docker-compose.run.yml and parse it with jsonnet

        vars_only: If False, do nothing, if True or non empty list, just dump the variables
        explain: Show the explainer and/or build diff

        """

        # 1. Prepare assemble context
        # -------------------
        sta = StackAssembler(parent=self, ident=f"{self.stack_name}")
        all_tags, jsonnet_lookup_dirs = self.get_tag_plan()
        self.var_manager = self.get_stack_vars(sta, all_tags, jsonnet_lookup_dirs)

        # 2. Prepare debugging tools
        # -------------------
        dump_payload_log = self.runtime.dump_payload_log
        self.dump_data_log = dump_payloads
        if explain and not vars_only:
            dumper = StackDumper(self.stack_dump_path, enabled=True)

        # 3. Prepare scopes
        # -------------------
        glob_vars = self.render_vars(scope="global", hint="global scoped variables")
        stack_vars = self.render_vars(
            scope="global,stack",
            parse=True,
            parse_vars=glob_vars,
            hint="stack scoped variables",
        )

        # 4. Intermediate debugging tools
        # -------------------
        self.log_extra_payloads(stack_vars, "Dump of docker environment vars")

        if explain and not vars_only:
            self.dump_diff_stage1(
                dumper, self.var_manager, glob_vars, stack_vars, all_tags
            )

        if vars_only:
            if explain:
                self.var_manager.explain(filter_vars=vars_only)

            out = dict(stack_vars)
            if isinstance(vars_only, list):
                out = {key: stack_vars.get(key) for key in vars_only}
            return out

        # 2. Build docker-compose
        # -------------------
        docker_run_payload = sta.assemble_docker_compose(
            all_tags,
            self.engine,
            env=stack_vars,
            dump_payload_log=dump_payload_log,
        )

        if explain:
            dumper.dump("1-docker-compose.yml", docker_run_payload, fmt="yaml")

        # 3. Assemble jsonnet tags
        # -------------------
        cand_index = (
            -2
        )  # We start at minus 2 here, because there is the 2 firsts tags are hidden
        tag_instances = []
        tag_names = []

        # Loop over each tags
        for cand in all_tags:
            cand_index += 1

            # 3.0 Init loop
            # --------------------

            # Check tag infos
            tag = cand.get("tag")
            tag_name = tag.name if tag else "_paasify"

            # Check conditions
            jsonnet_file = cand.get("jsonnet_file")
            if not jsonnet_file:
                continue

            # 3.1 Reload var context if overriden
            # --------------------
            tag_index = 0
            tag_inst = f"{tag_name}{tag_index}"
            while tag_inst in tag_instances:
                tag_index += 1
                tag_inst = f"{tag_name}{tag_index}"
            tag_instances.append(tag_inst)
            if tag_name not in tag_names:
                tag_names.append(tag_name)

            tag_suffix = f"{tag_index}" if tag_index != 0 else ""
            self.log.info(f"Apply jsonnet tag '{tag_inst}': {jsonnet_file}")

            # 3.2 Generate loop vars
            # --------------------
            scope = f"global,stack,tag_{tag_inst}"

            def scope_filter(var):
                # pylint: disable=cell-var-from-loop
                return var.scope in scope.split(",") and var.owner != tag_name

            loop_vars = self.render_vars(
                scope=scope_filter,
                parse=True,
                parse_vars=stack_vars,
                hint=f"jsonnet transform vars: {tag_inst}",
            )

            loop_vars.update(
                {
                    "tag_cand": cand_index,
                    "tag_index": tag_index,
                    "tag_instance": tag_inst,
                    "tag_suffix": tag_suffix,
                }
            )

            # Logging
            self.log_extra_payloads(
                loop_vars, f"Dump of vars before '{tag_inst}' jsonnet execution"
            )
            if explain:
                self.dump_diff_stage2(
                    dumper,
                    self.var_manager,
                    loop_vars,
                    scope_filter,
                    tag_inst,
                    cand_index,
                )

            # 3.3 Prepare jsonnet call
            # --------------------
            params = {
                "args": loop_vars,
                "docker_data": docker_run_payload,
            }
            docker_run_payload = sta.process_jsonnet_exec(
                jsonnet_file,
                "docker_transform",
                params,
                import_dirs=jsonnet_lookup_dirs,
            )
            if explain:
                dumper.dump(
                    f"2-{cand_index:02d}-{tag_inst}-out.yml",
                    docker_run_payload,
                    fmt="yml",
                )

        # 4. Write output file
        # -------------------

        if explain:
            dumper.dump("3-docker.yml", docker_run_payload, fmt="yml")

        # Prepare docker-file output directory
        if not os.path.isdir(self.stack_path):
            self.log.info(f"Create missing directory: {self.stack_path}")
            os.makedirs(self.stack_path)

        # Save the final docker-compose.run.yml file
        outfile = os.path.join(self.stack_path, "docker-compose.run.yml")
        self.log.info(f"Writing docker-compose file: {outfile}")
        output = to_yaml(docker_run_payload)
        write_file(outfile, output)

        if explain:
            dumper.show_diff()

        # 5. Prepare environment
        # -------------------
        # This is a first a basic implementation of apps volumes with permissions
        # TOFIX: Permission change will apply the same permissions on all volumes in a blind way
        # Ie: Permission for mysql containers is not the same as the app itself.
        volumes = docker_run_payload.get("volumes", {})
        uid = int(stack_vars.get("app_puid", "-1"))
        gid = int(stack_vars.get("app_pgid", "-1"))
        for vol_name, vol_def in volumes.items():
            driver = vol_def.get("driver")
            driver_opts = vol_def.get("driver_opts")
            if driver == "local" and driver_opts:
                device = driver_opts.get("device")
                if device and not os.path.exists(device):
                    self.log.info(
                        f"Create volume directory '{vol_name}' with owner '{uid}:{gid}': {device}"
                    )
                    os.makedirs(device)
                    os.chown(device, uid, gid)

    def explain_tags(self):
        "Explain hos tags are processed on stack"

        print(f"  Scanning stack plugins: {self.ident}")
        matches, jsonnet_lookup_dirs = self.get_tag_plan()

        # 0. Internal functions
        def list_items(items):
            "List items"
            _first = "*"
            # list_items(items)
            for cand in items:
                print(f"          {_first} {cand}")
                _first = "-"

        def list_jsonnet_files(items):
            "List jsonnet files"
            for match in items:

                tag = match.get("tag")
                if tag:
                    cand = match.get("jsonnet_file")
                    if cand:
                        print(f"        - {tag.name}")

        # 1. Show all match combinations
        for match in matches:

            tag = match.get("tag")

            if not tag:
                print("    Default config:")
                list_items(self.docker_candidates())
                print("    Tag config:")
                continue

            print(f"      tag: {tag.name}")

            if tag.docker_candidates:
                print("        Docker tags:")
                list_items(tag.docker_candidates)

            if tag.jsonnet_candidates:
                print("        Jsonnet tags:")
                list_items(tag.jsonnet_candidates)

        # 2. Show actual loading
        print("\n    Tag Loading Order:")

        # 2.1 Var loading
        print("      Loading vars:")
        list_jsonnet_files(matches)

        # 2.2 Tag loading
        print("      Loading Tags:")
        for match in matches:

            tag = match.get("tag")

            if not tag:
                cand = match.get("docker_file")
                print(f"        * base: {cand}")
                continue

            cand = match.get("docker_file")
            if cand:
                print(f"        - {tag.name}")

        # 2.3 Jsonnet loading
        print("      Loading jsonnet:")
        list_jsonnet_files(matches)

    def gen_doc(self, output_dir=None):
        "Generate documentation"

        matches, jsonnet_lookup_dirs = self.get_tag_plan()

        # 3. Show jsonschema
        print("\n    Plugins jsonschema:")
        for match in matches:

            tag = match.get("tag")
            if not tag:
                continue

            file = match.get("jsonnet_file")
            if not file:
                continue

            # Create output dir
            dest_dir = os.path.join(output_dir, tag.name)
            if not os.path.isdir(dest_dir):
                os.makedirs(dest_dir)

            print(f"        # {tag.name}: {file}")
            out = self.process_jsonnet(file, "metadata", None)
            tag_meta = out["metadata"]
            tag_schema = tag_meta.get("jsonschema")
            # pprint (tag_meta)
            if "jsonschema" in tag_meta:
                del tag_meta["jsonschema"]

            dest_schema = os.path.join(dest_dir, "jsonschema")
            if tag_schema:
                print(f"Generated jsonschema files in: {dest_schema}.[json|yml]")
                write_file(dest_schema + ".json", to_json(tag_schema))
                write_file(dest_schema + ".yml", to_yaml(tag_schema))

            # Create HTML documentation
            if ENABLE_JSON_SCHEMA:

                fname = "web.html"
                dest_html = os.path.join(dest_dir, fname)
                print(f"Generated HTML doc in: {dest_html}")
                config = GenerationConfiguration(
                    copy_css=True,
                    description_is_markdown=True,
                    examples_as_yaml=True,
                    footer_show_time=False,
                    expand_buttons=True,
                    show_breadcrumbs=False,
                )
                generate_from_filename(dest_schema + ".json", dest_html, config=config)

                # /schema_doc/paasify_yml_schema.html
                # /plugin_api_doc/{tag.name}/web.html
                markdown_doc = f"""
# {tag.name}

Documentationfor tag: `{tag.name}`

## Informations

``` yaml
{to_yaml(tag_meta)}
```

## Tag documentation

<iframe scrolling="yes" src="/plugins_apidoc/{tag.name}/{fname}" style="width: 100vw; height: 70vh; overflow: auto; border: 0px;">
</iframe>


                """
                dest_md = os.path.join(dest_dir, "markdown.md")
                write_file(dest_md, markdown_doc)
                print(f"Generated Markdown doc in: {dest_md}")


def stack_target(fn):
    "Decorator to magically find the correct stack to apply"

    @wraps(fn)
    def wrapper(self, *args, stacks=None, stack_names=None, **kwargs):
        "Decorator to magically find the correct stack to apply"

        # Inteligently guess wich stack to use
        if not stacks:
            stacks = []

            if isinstance(stack_names, str):
                stack_names = [name.strip("/") for name in stack_names.split(",")]

            sub_dir = self.get_parent().runtime.sub_dir
            if not stack_names:

                if sub_dir:

                    # Use current dir stacks if in subdir
                    stack_path = sub_dir.split(os.path.sep)[0]
                    stacks = [
                        stack
                        for stack in self.get_children()
                        if stack_path == stack.stack_dir
                    ]
                    if stacks:
                        self.log.debug(
                            f"Use stack {stack_path} from current working directory"
                        )

                else:
                    # Use all stacks is set to None
                    self.log.debug("Use all stacks")
                    stacks = self.get_children()

            else:
                # Loop over specified list of tasks
                assert isinstance(stack_names, list), f"Got: {stack_names}"
                stack_names = [name for name in stack_names if name]
                stacks = [
                    stack
                    for stack in self.get_children()
                    if stack.stack_name in stack_names
                    or stack.stack_path in stack_names
                ]
                self.log.debug(
                    f"Requested stacks: {stack_names}, matched stacks: {stacks}"
                )

                # Assert all stacks has been addressed
                if len(stack_names) != len(stacks):
                    missing_stacks = list(stack_names)
                    for stack in stacks:
                        missing_stacks.remove(stack.stack_name)
                    raise error.StackNotFound(
                        f"This stack is not defined in config: {','.join(missing_stacks)}"
                    )

        # Clean decorator argument
        if "stacks_names" in kwargs:
            del kwargs["stacks_names"]

        # Last sanity tests
        if len(self.get_children()) < 1:
            raise error.StackNotFound(
                "There are no stacks configured yet for this project, please edit your paasify.yml config"
            )
        if len(stacks) < 1:
            raise error.StackNotFound(
                f"This stack is not defined in config: {stack_names}"
            )
        assert isinstance(stacks, list), f"Got: {stacks}"

        # Last user report
        stacks_names = ",".join([stack.stack_name for stack in stacks])
        self.log.info(f"Running command '{fn.__name__}' on stacks: {stacks_names}")
        return fn(self, *args, stacks=stacks, **kwargs)

    return wrapper


class StackManager(NodeList, PaasifyObj):
    "Manage a list of stacks"

    conf_schema = {
        # "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "Paasify Stack configuration",
        "description": "Stacks are defined in a list of objects",
        "type": "array",
        "default": [],
        "items": Stack.conf_schema,
    }

    conf_children = Stack
    ident = "main"

    def node_hook_final(self):
        "Enable CLI logging and validate config"

        # Safety checks
        dup_names = []
        dup_dirs = []
        curr_index = -1
        for stack in self.get_children():
            curr_index += 1
            stack_name = stack.stack_name

            # Check for duplicates names
            if stack_name in dup_names:
                index = dup_names.index(stack_name)
                raise error.ProjectInvalidConfig(
                    f"Cannot have duplicate stack names, stacks {index} and {curr_index} share the same name: '{stack_name}'"
                )
            dup_names.append(stack_name)

            # Check for duplicates dirs
            stack_dir = stack.stack_dir
            if stack_dir in dup_dirs:
                index = dup_dirs.index(stack_dir)
                raise error.ProjectInvalidConfig(
                    f"Cannot have duplicate stack dir, stacks {index} and {curr_index} share the same dir: '{stack_dir}'"
                )
            dup_dirs.append(stack_dir)

        # Notice user because this is weird
        children = self.get_children()
        if not children:
            self.log.warning("No stacks found for this project!")

    # Stack management API
    # ======================

    def list_stacks(self):
        "Get stacks children (deprecated)"
        return self.get_children()

    def get_stacks_attr_ident(self):
        "List stack per idents"
        return [x.ident for x in self.get_children()]

    def get_stacks_attr(self, attr="ident"):
        "List stacks by attribute"
        return [getattr(x, attr) for x in self.get_children()]

    def get_stacks_obj(self, attr=None, values=None):
        """
        Get stack instance matching in values

        If attr or value is None, return all instances
        Values must be an array of valid vallues.
        """

        sub_dir = self._node_parent.runtime.sub_dir
        if values is None and sub_dir:
            # attr = path
            values = sub_dir.split(os.path.sep)[0]
            print(f"Automatch subdir: {values}")

        if isinstance(attr, str) and values is not None:
            if not isinstance(values, list):
                values = [values]
            result = [
                stack for stack in self.get_children() if getattr(stack, attr) in values
            ]
            return result

        return self.get_children()

    # Command Base API
    # ======================

    @stack_target
    def cmd_stack_assemble(self, stacks=None, vars_only=False, explain=False):
        "Assemble a stack"

        self.log.info("Asemble stacks:")
        for stack in stacks:
            self.log.notice(f"Assemble stack: {stack.stack_name}")
            stack.assemble(vars_only=vars_only, explain=explain)

    @stack_target
    def cmd_stack_up(self, stacks=None):
        "Start a stack"

        self.log.info("Start stacks:")
        for stack in stacks:
            self.log.notice(f"  Start stack: {stack.stack_name}")
            stack.engine.up(**stack.default_sh_args)

    @stack_target
    def cmd_stack_down(self, stacks=None, ignore_errors=False):
        "Stop a stack"

        stacks = list(stacks)
        stacks.reverse()
        self.log.info("Stop stacks:")
        for stack in stacks:
            self.log.notice(f"  Stop stack: {stack.stack_name}")
            try:
                stack.engine.down(**stack.default_sh_args)
            except error.DockerCommandFailed:
                if not ignore_errors:
                    raise
                self.log.debug(
                    f"Ignoring stop failure in case of recreate for stack: {stack.stack_name}"
                )

    @stack_target
    def cmd_stack_ps(self, stacks=None):
        "List stacks process"

        if len(stacks) < 1:
            self.log.notice("  No process founds")
            return

        # self.log.notice("List of processes:")
        for stack in stacks:
            # self.log.notice(f"Process of stack: {stack.stack_name}")
            stack.engine.ps()

    # Shortcuts
    # ======================
    @stack_target
    def cmd_stack_apply(self, stacks=None):
        "Apply a stack"

        self.log.notice("Apply stacks")
        self.cmd_stack_assemble(stacks=stacks)
        self.cmd_stack_up(stacks=stacks)
        self.log.notice("Stack has been applied")

    @stack_target
    def cmd_stack_recreate(self, stacks=None):
        "Recreate a stack"

        self.log.notice("Recreate stacks")
        self.cmd_stack_down(stacks=stacks, ignore_errors=True)
        self.cmd_stack_assemble(stacks=stacks)
        self.cmd_stack_up(stacks=stacks)
        self.log.notice("Stack has been recreated")

    # Other commands
    # ======================

    def cmd_stack_ls(self, stacks=None):
        "List command to stacks"

        for stack in self.get_children():
            stack_app = stack.app.app if stack.app else None
            print(f"  - {stack.stack_name}:")
            print(f"      app: {stack_app}")
            print(f"      path: {stack.stack_path}")

    @stack_target
    def cmd_stack_vars(self, stacks=None, vars_=None, explain=False):
        "Show vars of stack"

        if isinstance(vars_, list):
            self.log.info(f"Restrict output to variables: {','.join(vars_)}")

        for stack in stacks:
            self.log.notice(f"Get stack vars: {stack.stack_name}")
            ret = stack.assemble(vars_only=vars_ or True, explain=explain)
            # TODO: PAtch cafram to support ordered yaml output !
            # See: https://github.com/barbu-it/cafram/blob/main/cafram/utils.py#L261
            # See: https://stackoverflow.com/questions/40226610/ruamel-yaml-equivalent-of-sort-keys
            if ret:
                pprint(ret)
            # print (to_yaml(ret))

    @stack_target
    def cmd_stack_explain(self, stacks=None):
        "Show informations on project plugins"

        for stack in stacks:
            stack.explain_tags()

        # if isinstance(mode, str):
        #     dst_path = mode
        #     self.log.notice("Generate documentation in dir:", dst_path)
        #     for stack in self.get_children():
        #         stack.gen_doc(output_dir=dst_path)

    @stack_target
    def cmd_stack_logs(self, stacks=None, follow=False):
        "Display stack/services logs"

        if follow and len(stacks) > 1:
            self.log.warning(
                "Disabling log following as it's not possible on more tha one stack"
            )
            follow = None

        for stack in stacks:
            self.log.notice(f"Logs of stack: {stack.stack_name}")
            stack.engine.logs(follow)



# -*- coding: utf-8 -*-
"Paasify Vars Libary"

# pylint: disable=logging-fstring-interpolation

import os
import logging


from pprint import pprint  # noqa: F401

import anyconfig

import paasify.errors as error
from paasify.common import StringTemplate, uniq
from paasify.framework import PaasifyObj

_log = logging.getLogger()


class Variable(PaasifyObj):
    "A simple variable with metadata"

    def __init__(self, parent=None, ident=None, payload=None):

        # Mandatory
        self.name = payload["name"]
        self.value = payload["value"]

        # Required
        self.index = payload["index"]
        self.file = payload["file"]
        self.scope = payload["scope"]
        self.parse_order = payload["parse_order"]

        # WIP
        self.source = payload.get("source")
        self.kind = payload.get("kind")
        self.owner = payload.get("owner")

        # Call parents
        self.ident = (
            f"{self.name}={self.value} ({self.index},{self.kind},{self.source})"
        )
        PaasifyObj.__init__(self, parent, ident)

    # def get_value(self):
    #    return {
    #            "name": None,
    #            }


class VarMgr(PaasifyObj):
    """
    This class manage a list of variables (PaasifyConfigVar), but it keep
    the adding order.

    Args:
        PaasifyObj (_type_): _description_

    Raises:
        error.ProjectInvalidConfig: _description_

    Returns:
        _type_: _description_
    """

    _vars = []

    def __init__(self, *arsg, **kwargs):

        self._vars = []
        super().__init__(*arsg, **kwargs)
        # self.set_logger("paasify.cli.vars")
        self.index = 0

    def add_vars(self, payload, **kwargs):
        "Add any kind of variables to the stack"

        # TODO: Sanity avoid duplicate piorities, should not be done this way, if still required
        prio = kwargs.get("parse_order")
        if prio:
            known_prios = uniq([var.parse_order for var in self._vars])
            if prio in known_prios:
                pprint([var for var in self._vars if var.parse_order == prio])
                assert False, f"Got duplicate priority: {prio}, found: {known_prios}"

        if isinstance(payload, dict):
            for name, var in payload.items():
                self.add_var(name, var, **kwargs)

        elif isinstance(payload, list):
            for var in payload:
                self.add_var(var, **kwargs)
        else:
            assert (
                False
            ), f"Object is not supported, expected dict or list, got: {payload}"

    def add_var(self, key, value=None, **kwargs):
        """Add a list/dict of vars into varmanager. You can override source/file/owner/kind.

        object can be: Variable (as key or value)
        object can be: key value
        """

        obj = None
        if isinstance(key, Variable):
            obj = key
        elif isinstance(value, Variable):
            obj = value

        if obj:
            # Assign overrides
            for name, val in kwargs.items():
                try:
                    getattr(obj, name)
                except AttributeError:
                    assert False, "BUG HERE !!!"
                setattr(obj, name, val)
        else:
            self.index += 1
            payload = dict(kwargs)
            payload.update(
                {
                    "name": key,
                    "value": value,
                    "index": self.index,
                }
            )
            obj = Variable(parent=self, ident="StackVar", payload=payload)

        self._vars.append(obj)

    def add_vars_from_lookup(
        self, lookup, parse_order, fail_on_missing=False, **override
    ):
        """Process yml vars from a tag_list"""

        vars_cand = lookup.match(fail_on_missing=fail_on_missing)
        for cand in vars_cand:

            # Parse file
            cand_file = cand["match"]
            ac_parser = "yaml"
            if cand_file.endswith("json"):
                ac_parser = "json"
            elif cand_file.endswith("toml"):
                ac_parser = "toml"
            self.log.info(f"Process {ac_parser} var file: {cand_file}")
            conf = anyconfig.load(cand_file, ac_parser=ac_parser)
            assert isinstance(conf, dict)

            config = {
                "kind": cand["kind"] + f"_{ac_parser}",
                "source": "core",
                "file": cand_file,
                "owner": cand["owner"],
                "parse_order": parse_order,
            }
            config.update(override)

            # Append to config
            self.add_vars(conf, **config)
            parse_order += 1

    # Vars selector
    # ===========================

    def select(self, func):
        "Allow to select vars on any attributes with a function"
        assert callable(func), f"Wrong argument type, expected a callable, got: {func}"
        result = [var for var in self._vars if func(var)]
        return result

    # Templating processors
    # ===========================

    # TODO: This should be a simple function, not a staticmethod
    @staticmethod
    def _render_env_sorter(var):
        "Helper function to sort vars per parse_order"
        return var.parse_order

    def render_env(
        self, parse=False, parse_vars=None, select=None, hint=None, skip_undefined=False
    ):
        """Return environment context, with eventually parsed variables"""

        # Prepare args
        parse_vars = parse_vars or {}
        selection = []
        if select:
            selection = list(self.select(select))
        else:
            selection = list(self._vars)

        # Prepare result
        env = {
            var.name: var.value
            for var in sorted(selection, key=self._render_env_sorter)
        }
        if not parse:
            self.log.debug(f"Fetch vars: {hint}")
            return env

        # Parse results
        self.log.debug(f"Parse vars, {hint}")
        result = {}
        result2 = dict(env)
        result2.update(parse_vars)
        for var in sorted(selection, key=self._render_env_sorter):
            value = self.template_value(
                var.value, result2, hint=var.name, skip_undefined=skip_undefined
            )
            result[var.name] = value
            result2[var.name] = value

        return result

    def resolve_dyn_vars(self, tpl, env, hint=None):
        "Resolver environment and secret vars"

        env = dict(env)
        var_list = tpl.get_identifiers()
        line = tpl.template

        for var in var_list:
            if var.startswith("_env_"):
                name = var[5:]
                value = os.environ.get(name)
                msg = f"Fetching environment value for: {hint}: {line} ({name}={value})"
                self.log.info(msg)
                env[var] = value
            elif var.startswith("_secret_"):
                msg = f"Support for secrets is not implemented yet: {hint}: {line}"
                self.log.warning(msg)
                env[var] = var
                # raise NotImplementedError(msg)

        return env

    def template_value(self, value, env, hint=None, skip_undefined=False):
        "Render a string with template engine"

        if not isinstance(value, str):
            return value

        # Resolve dynamic vars
        tpl = StringTemplate(value)
        # env = self.resolve_dyn_vars(tpl, env, hint=hint)

        # pylint: disable=broad-except
        try:
            old_value = value
            value = tpl.substitute(**env)
            if old_value != value:
                self.log.trace(
                    f"Transformed template var {hint}: {old_value} => {value}"
                )

        except KeyError as err:
            if not skip_undefined:
                msg = f"Variable {err} is not defined in variable '{hint}': '{value}'"
                # self.log.warning(msg)
                raise error.UndeclaredVariable(msg) from KeyError

        except ValueError:
            self.log.debug(
                f"Could not parse variable: {hint}='{value}', forwarding to docker compose"
            )

        return value

    # Debug and troubleshoot
    # ===========================

    def dump(self, *args, **kwargs):
        """
        Dump and explain the current variable stack
        """
        super().dump(*args, **kwargs)
        print("  Vars:")
        print("  -----------------")
        filter_vars = kwargs.get("filter_vars")
        filter_vars = None if filter_vars is True else filter_vars
        self.explain()

    def explain(self, scope=None, as_string=False, filter_vars=None):
        """
        Explain the current variable stack
        """

        # header
        payload = {
            "name": "NAME",
            "value": "VALUE",
            "kind": "KIND",
            "source": "SOURCE",
            "file": "FILE",
            "scope": "SCOPE",
            "owner": "OWNER",
            "index": "INDEX",
            "parse_order": "PARSE_ORDER",
        }
        header = [Variable(payload=payload)]

        # Prepare args
        selection = []
        if scope:
            selection = list(self.select(scope))
        else:
            selection = list(self._vars)

        # Loop over all vars
        ellipsis = "..."
        max_ = 60
        result = []
        if not isinstance(filter_vars, list):
            filter_vars = None
        for var in header + sorted(selection, key=self._render_env_sorter):

            if filter_vars and var.name not in filter_vars:
                continue

            # TODO: There is a lib for that !
            ell = ellipsis if len(str(var.value)) > max_ else ""
            value = str(var.value)[: max_ - len(ell)] + ell
            while len(value) < max_:
                add = max_ - len(value)
                add = " " * add
                value += add

            # txt = f"{var.parse_order:<6}{var.name:40}: {value} {var.scope:20} {var.file}"
            # txt = f"{var.parse_order:<6}{var.name:30}: {value:52} {var.scope:20}{var.owner:20}{var.kind:20}"

            # What user want to see
            txt = f"{var.parse_order:<6}{var.name:30}: {value:52} {var.scope:20}  {var.file}"
            result.append(txt)

        result = "\n".join(result)
        if as_string:
            return result

        print(result)
